import asyncio
from functools import update_wrapper
from typing import AsyncIterator, Callable, Any, Union, AsyncGenerator, Generator, Iterator, Awaitable

from .apraw_base import aPRAWBase
from ...utils import ExponentialCounter

# I know, I know, this code is cursed
SYNC_OR_ASYNC_ITERABLE = Union[
    Callable[[Any, int, Any], Union[Awaitable[Union[AsyncIterator[aPRAWBase], Iterator[aPRAWBase]]], Union[
        AsyncIterator[aPRAWBase], Iterator[aPRAWBase]]]], AsyncGenerator[aPRAWBase, None], Generator[
        aPRAWBase, None, None]]


class Streamable:
    """
    A decorator to make functions returning a generator streamable.

    Members
    -------
    max_wait: int
        The maximum amount of seconds to wait before repolling the function.
    attribute_name: str
        The attribute name to use as a unique identifier for returned objects.
    """

    @classmethod
    def streamable(cls, func: SYNC_OR_ASYNC_ITERABLE = None, max_wait: int = 16, attribute_name: str = "fullname"):
        if func:
            return Streamable(func)
        else:
            def wrapper(
                    _func: SYNC_OR_ASYNC_ITERABLE):
                return Streamable(_func, max_wait, attribute_name)

            return wrapper

    def __init__(self, func: SYNC_OR_ASYNC_ITERABLE, max_wait: int = 16, attribute_name: str = "fullname"):
        """
        Create an instance of the streamable object.

        Parameters
        ----------
        func: Callable[[Any, int, Any], AsyncIterator[Any]]
            The function returning an asynchronous iterator.
        max_wait: int
            The maximum amount of seconds to wait before repolling the function.
        attribute_name: str
            The attribute name to use as a unique identifier for returned objects.
        """
        self._func = func
        update_wrapper(self, func)

        self.max_wait = max_wait
        self.attribute_name = attribute_name

    def __get__(self, instance: Any, owner: Any):
        """
        Allow streamable to access its top-level object instance to forward to functions later.
        """
        self.instance = instance
        return self

    async def __call__(self, *args, **kwargs):
        """
        Make streamable callable to return result of decorated function.
        """
        if hasattr(self._func, "__call__"):
            func_args = (self.instance, *args) if hasattr(self, "instance") else args
            if asyncio.iscoroutinefunction(self._func):
                iterable = await self._func(*func_args, **kwargs)
            else:
                iterable = self._func(*func_args, **kwargs)
        else:
            iterable = self._func

        if hasattr(iterable, "__aiter__"):
            async for item in iterable:
                yield item
        else:
            for item in iterable:
                yield item

    async def stream(self, skip_existing: bool = False, *args, **kwargs):
        r"""
        Call the stream method on the decorated function.

        Parameters
        ----------
        skip_existing: bool
            Whether items found before the function call should be returned as well.
        kwargs: \*\*Dict
            ``kwargs`` to be passed on to the function.

        Yields
        ------
        item: aPRAWBase
            The item retrieved by the function in chronological order.
        """
        counter = ExponentialCounter(self.max_wait)
        seen_attributes = list()

        while True:
            found = False
            items = [i async for i in self(100, *args, **kwargs)]
            for item in reversed(items):
                attribute = getattr(item, self.attribute_name)
                if attribute in seen_attributes:
                    continue

                if len(seen_attributes) >= 301:
                    seen_attributes = seen_attributes[1:]

                seen_attributes.append(attribute)
                found = True
                if not skip_existing:
                    yield item

            skip_existing = False

            if found:
                wait = counter.reset()
            else:
                wait = counter.count()

            await asyncio.sleep(wait)



from ...const import API_PATH


class HideableMixin:
    """
    Mixin for hideable objects.
    """

    async def hide(self):
        """
        Hide the item.

        Returns
        -------
        resp: Dict
            The API response JSON.
        """
        return await self.reddit.post_request(API_PATH["post_hide"], id=self.fullname)

    async def unhide(self):
        """
        Unhide the item.

        Returns
        -------
        resp: Dict
            The API response JSON.
        """
        return await self.reddit.post_request(API_PATH["post_unhide"], id=self.fullname)



from typing import TYPE_CHECKING, AsyncIterator, Awaitable, List

from .apraw_base import aPRAWBase
from ..reddit.comment import Comment
from ..reddit.submission import Submission
from ..subreddit.moderation import ModAction
from ..subreddit.subreddit import Subreddit

if TYPE_CHECKING:
    from ...reddit import Reddit


class ListingGenerator(AsyncIterator):
    """
    The model to request listings from Reddit.

    Members
    -------
    reddit: Reddit
        The :class:`~apraw.Reddit` instance with which requests are made.
    endpoint: str
        The endpoint to make requests on.
    max_wait: int
        The maximum amount of seconds to wait before re-requesting in streams.
    kind_filter:
        Kinds to return if given, otherwise all are returned.
    subreddit: Subreddit
        The subreddit to inject as a dependency into items if given.

    .. note::
        ListingGenerator will automatically make requests until none more are found or the limit has been reached.
    """

    def __init__(self, reddit: 'Reddit', endpoint: str, limit: int = 100, subreddit: Subreddit = None,
                 kind_filter: List[str] = None, **kwargs):
        r"""
        Create a ``ListingGenerator`` instance.

        Parameters
        ----------
        reddit: Reddit
            The :class:`~apraw.Reddit` instance with which requests are made.
        endpoint: str
            The endpoint to make requests on.
        limit: int
            The maximum number of items to fetch.
        max_wait: int
            The maximum amount of seconds to wait before re-requesting in streams.
        subreddit: Subreddit
            The subreddit to inject as a dependency into items if given.
        kwargs: \*\*Dict
            Query parameters to append to the request URL.
        """
        self.reddit = reddit
        self.endpoint = endpoint
        self.limit = limit if limit else 1024
        self.subreddit = subreddit
        self.params = {**kwargs, "limit": self.limit}
        self.listing = None
        self.kind_filter = kind_filter
        self._yielded = 0

    def __aiter__(self) -> AsyncIterator[aPRAWBase]:
        """
        Permit ListingGenerator to operate as an iterator.

        Returns
        -------
        self: ListingGenerator
            The iterator.
        """
        return self

    async def __anext__(self) -> Awaitable[aPRAWBase]:
        """
        Permit ListingGenerator to operate as a generator.

        Returns the next item in the listing.

        Returns
        -------
        subreddit: Subreddit
            The subreddit found in the listing.
        comment: Comment
            The comment found in the listing.
        submission: Submission
            The submission found in the listing.
        mod_action: ModAction
            The mod action found in the listing.
        wikipage_revision: WikipageRevision
            The wikipage revision found in the listing.
        item: aPRAWBase
            A model of the item's data if kind couldn't be identified.
        """
        if self._yielded >= self.limit:
            raise StopAsyncIteration()

        if self.listing is None:
            await self._next_batch()

        try:
            item = next(self.listing)
            self._yielded += 1
            return item
        except StopIteration:
            if self._yielded < self.limit:
                await self._next_batch()
            else:
                raise StopAsyncIteration()

    async def _next_batch(self):
        """
        Retrieve the next batch of items and store them in a :class:`~apraw.models.Listing`.
        """
        kwargs = {**self.params}

        if self.listing:
            kwargs["after"] = self.listing.last.fullname

        self.listing = await self.reddit.get_listing(self.endpoint, self.subreddit, self.kind_filter, **kwargs)

        if len(self.listing) <= 0:
            raise StopAsyncIteration()



from ...const import API_PATH


class NSFWableMixin:
    """
    Mixin for NSFWable objects.
    """

    async def mark_nsfw(self):
        """
        Mark the item as NSFW.

        Returns
        -------
        resp: Dict
            The API response JSON.
        """
        return await self.reddit.post_request(API_PATH["post_marknsfw"], id=self.fullname)

    async def unmark_nsfw(self):
        """
        Unmark the item as NSFW.

        Returns
        -------
        resp: Dict
            The API response JSON.
        """
        return await self.reddit.post_request(API_PATH["post_unmarknsfw"], id=self.fullname)



from enum import Enum
from typing import TYPE_CHECKING, NewType

from ...const import API_PATH
from .apraw_base import aPRAWBase

if TYPE_CHECKING:
    from ...reddit import Reddit


class DistinguishmentOption(Enum):
    YES = "yes"
    NO = "no"
    ADMIN = "admin"
    SPECIAL = "special"


DISTINGUISHMENT_OPTIONS = NewType(
    "DistinguishmentOptions",
    DistinguishmentOption)


class ItemModeration:
    """
    A helper class to moderate comments, submissions and modmail.

    Members
    -------
    reddit: Reddit
        The :class:`~apraw.Reddit` instance with which requests are made.
    fullname: str
        The ID prepended with the kind of the item this helper belongs to.
    """

    def __init__(self, reddit: 'Reddit', item: aPRAWBase):
        """
        Create an instance of ``ItemModeration``.

        Parameters
        ----------
        reddit : Reddit
            The :class:`~apraw.Reddit` instance with which requests are made.
        item : aPRAWBase
            The item this helper performs requests for.
        """
        self.reddit = reddit
        self.fullname = item.fullname

    async def approve(self):
        """
        Approve the Reddit item.

        Returns
        -------
        resp: Dict
            The API response JSON.
        """
        return await self.reddit.post_request(API_PATH["mod_approve"], id=self.fullname)

    async def remove(self):
        """
        Remove the Reddit item.

        Returns
        -------
        resp: Dict
            The API response JSON.
        """
        return await self.reddit.post_request(API_PATH["mod_remove"], id=self.fullname)

    async def distinguish(self,
                          how: DISTINGUISHMENT_OPTIONS = "yes",
                          sticky: bool = False):
        """
        Distinguish the Reddit item.

        Parameters
        ----------
        how : "yes" or "no" or "admin" or "special"
            The type of distinguishment to be added to the item.
        sticky : bool, optional
            Whether the item should be stickied.

        Returns
        -------
        resp: Dict
            The API response JSON.
        """
        return await self.reddit.post_request(API_PATH["mod_distinguish"],
                                              id=self.fullname, how=how, sticky=sticky)

    async def undistinguish(self):
        """
        Undistinguish the Reddit item.

        Returns
        -------
        resp: Dict
            The API response JSON.
        """
        return await self.reddit.post_request(API_PATH["mod_distinguish"],
                                              id=self.fullname, how="no", sticky=False)

    async def ignore_reports(self):
        """
        Ignore reports on the Reddit item.

        Returns
        -------
        resp: Dict
            The API response JSON.
        """
        return await self.reddit.post_request(API_PATH["mod_ignore_reports"], id=self.fullname)

    async def unignore_reports(self):
        """
        Unignore previously ignored reports on the Reddit item.

        Returns
        -------
        resp: Dict
            The API response JSON.
        """
        return await self.reddit.post_request(API_PATH["mod_unignore_reports"], id=self.fullname)


class PostModeration(ItemModeration):
    """
    A helper class to moderate comments and submissions.

    Members
    -------
    reddit: Reddit
        The :class:`~apraw.Reddit` instance with which requests are made.
    fullname: str
        The ID prepended with the kind of the item this helper belongs to.
    """

    def __init__(self, reddit, item):
        """
        Create an instance of ``PostModeration``.

        Parameters
        ----------
        reddit : Reddit
            The :class:`~apraw.Reddit` instance with which requests are made.
        item : aPRAWBase
            The item this helper performs requests for.
        """
        super().__init__(reddit, item)

    async def lock(self):
        """
        Lock the item from further replies.

        Returns
        -------
        resp: Dict
            The API response JSON.
        """
        return await self.reddit.post_request(API_PATH["mod_lock"], id=self.fullname)

    async def unlock(self):
        """
        Unlock the item from further replies.

        Returns
        -------
        resp: Dict
            The API response JSON.
        """
        return await self.reddit.post_request(API_PATH["mod_unlock"], id=self.fullname)



from typing import Dict, Any, TYPE_CHECKING

from ..helpers.apraw_base import aPRAWBase
from ..mixins.author import AuthorMixin
from ..mixins.replyable import ReplyableMixin
from ..mixins.subreddit import SubredditMixin

if TYPE_CHECKING:
    from ...reddit import Reddit


class Message(aPRAWBase, SubredditMixin, AuthorMixin, ReplyableMixin):

    def __init__(self, reddit: 'Reddit', data: Dict[str, Any]):
        super().__init__(reddit, data, reddit.message_kind)



from .helpers.apraw_base import aPRAWBase
from .helpers.generator import ListingGenerator
from .helpers.item_moderation import ItemModeration, PostModeration
from .helpers.streamable import Streamable
from .reddit.comment import Comment
from .reddit.listing import Listing
from .reddit.message import Message
from .reddit.more_comments import MoreComments
from .reddit.redditor import Redditor
from .reddit.submission import Submission
from .subreddit.moderation import ModAction
from .subreddit.modmail import ModmailConversation, ModmailMessage, SubredditModmail
from .subreddit.subreddit import Subreddit
from .subreddit.wiki import WikipageRevision
from .user import AuthenticatedUser, Karma, User



from ...const import API_PATH


class SpoilerableMixin:
    """
    Mixin for spoilerable objects.
    """

    async def spoiler(self):
        """
        Mark the item as a spoiler.

        Returns
        -------
        resp: Dict
            The API response JSON.
        """
        return await self.reddit.post_request(API_PATH["post_spoiler"], id=self.fullname)

    async def unspoiler(self):
        """
        Unmark the item as a spoiler.

        Returns
        -------
        resp: Dict
            The API response JSON.
        """
        return await self.reddit.post_request(API_PATH["post_unspoiler"], id=self.fullname)



from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict

from ...utils import prepend_kind, snake_case_keys

if TYPE_CHECKING:
    from ...reddit import Reddit


class aPRAWBase:
    """
    The base class for Reddit models.

    The ``aPRAWBase`` class stores data retrieved by the endpoints and automatically assigns it as attributes.
    Specific information about the aforementioned attributes can be found in the respective implementations such as :class:`~apraw.models.Comment`.

    Members
    -------
    kind: str
        The item's kind / type.
    """

    ID_ATTRIBUTE = "id"

    def __init__(self, reddit: 'Reddit', data: Dict[str, Any] = None, kind: str = ""):
        """
        Initialize the base information.

        Parameters
        ----------
        reddit: Reddit
            The :class:`~apraw.Reddit` instance with which requests are made.
        data: Dict
            The data obtained from the /about endpoint.
        """
        self._reddit = reddit
        self._data = data
        self._data_attrs = set()
        self.kind = kind

        if data:
            self._update(data)

    def _update(self, data: Dict[str, Any]):
        """
        Update the base with new information.

        Parameters
        ----------
        data: Dict
            The data obtained from the /about endpoint.
        """
        self._data = data

        d = snake_case_keys(data)
        for key in d:
            if not hasattr(self, key):
                self._data_attrs.add(key)
                setattr(self, key, d[key])
            elif key in self._data_attrs:
                setattr(self, key, d[key])

        if "created_utc" in data:
            self.created_utc = datetime.utcfromtimestamp(data["created_utc"])

    async def fetch(self):
        """
        Fetch this item's information from a suitable API endpoint.

        Returns
        -------
        self: aPRAWBase
            The updated model.
        """
        raise NotImplementedError

    def __repr__(self):
        """
        Get a representational string for this model following the pattern ``<{class} {id_attribute}='{id}'>``.

        Returns
        -------
        repr: string
            A printable representational string for this model.
        """
        return f"<{self.__class__.__name__} {self.ID_ATTRIBUTE}='{self._data[self.ID_ATTRIBUTE]}'>"

    @property
    def fullname(self):
        return prepend_kind(self._data[self.ID_ATTRIBUTE], self.kind) if self.kind else self._data[self.ID_ATTRIBUTE]



from typing import TYPE_CHECKING, Dict

from ..helpers.apraw_base import aPRAWBase
from ..helpers.streamable import Streamable
from ...const import API_PATH

if TYPE_CHECKING:
    from ..subreddit.subreddit import Subreddit
    from ...reddit import Reddit


class Redditor(aPRAWBase):
    """
    The model representing Redditors.

    Members
    -------
    reddit: Reddit
        The :class:`~apraw.Reddit` instance with which requests are made.
    data: Dict
        The data obtained from the /about endpoint.
    kind: str
        The item's kind / type.
    subreddit: Sureddit
        An instance of :class:`~apraw.models.Subreddit` for the Redditor's profile subreddit.

    **Typical Attributes**

    This table describes attributes that typically belong to objects of this
    class. Attributes are dynamically provided by the :class:`~apraw.models.aPRAWBase` class
    and may vary depending on the status of the response and expected objects.

    ======================== ================================================================
    Attribute                Description
    ======================== ================================================================
    ``comment_karma``        The amount of comment karma the Redditor has obtained.
    ``created_utc``          The date on which the Redditor was created in UTC ``datetime``.
    ``created``              The timestamp of when the Redditor was created.
    ``has_verified_email``   Whether the Redditor has a verified email address.
    ``icon_img``             A URL to the Redditor's icon image if applicable.
    ``id``                   The Redditor's ID (without kind).
    ``is_employee``          Whether the Redditor is a Reddit employee.
    ``is_friend``            Whether the Redditor has been added as a friend.
    ``is_gold``              Whether the Redditor is a Reddit gold member.
    ``is_mod``               Whether the Redditor is a moderator in a subreddit.
    ``is_suspended``         Whether the Redditor has been suspended.
    ``link_karma``           The amount of link karma the Redditor has obtained.
    ``name``                 The Redditor's username.
    ``pref_show_snoovatar``  Whether to show the Redditor's Snoovatar in place of their icon.
    ``verified``             Whether the Redditor is verified.
    ======================== ================================================================

    .. warning::
        Suspended Redditors only return ``is_suspended`` and ``name``.

    """

    def __init__(self, reddit: 'Reddit', data: Dict):
        """
        Create a Redditor instance.

        Parameters
        ----------
        reddit: Reddit
            The :class:`~apraw.Reddit` instance with which requests are made.
        data: Dict
            The data obtained from the /about endpoint.
        """
        super().__init__(reddit, data, reddit.account_kind)

        self.reddit = reddit
        self.data = data

        self.is_suspended = "is_suspended" not in data or not data["is_suspended"]

        if "subreddit" in data and data["subreddit"]:
            sub = data["subreddit"]
            sub["id"] = sub["name"].replace("t5_", "")
            if "created_utc" not in sub:
                sub["created_utc"] = data["created_utc"]
            from apraw.models.subreddit.subreddit import Subreddit
            self.subreddit = Subreddit(self.reddit, sub)
        else:
            self.subreddit = None

    async def fetch(self):
        """
        Fetch this item's information from a suitable API endpoint.

        Returns
        -------
        self: Redditor
            The updated ``Redditor``.
        """
        resp = await self._reddit.get_request(API_PATH["user_about"].format(user=self._data["username"]))
        self._update(resp["data"])
        return self

    @Streamable.streamable
    def comments(self, *args, **kwargs):
        r"""
        Returns an instance of :class:`~apraw.models.ListingGenerator` mapped to fetch the Redditor's comments.

        .. note::
            This listing can be streamed doing the following:

            .. code-block:: python3

                for comment in redditor.comments.stream():
                    print(comment)

        Parameters
        ----------
        kwargs: \*\*Dict
            :class:`~apraw.models.ListingGenerator` ``kwargs``.

        Returns
        -------
        generator: ListingGenerator
            A :class:`~apraw.models.ListingGenerator` mapped to fetch the Redditor's comments.
        """
        from ..helpers.generator import ListingGenerator
        return ListingGenerator(self.reddit, API_PATH["user_comments"].format(user=self), *args, **kwargs)

    @Streamable.streamable
    def submissions(self, *args, **kwargs):
        r"""
        Returns an instance of :class:`~apraw.models.ListingGenerator` mapped to fetch the Redditor's submissions.

        .. note::
            This listing can be streamed doing the following:

            .. code-block:: python3

                for comment in redditor.submissions.stream():
                    print(comment)

        Parameters
        ----------
        kwargs: \*\*Dict
            :class:`~apraw.models.ListingGenerator` ``kwargs``.

        Returns
        -------
        generator: ListingGenerator
            A :class:`~apraw.models.ListingGenerator` mapped to fetch the Redditor's submissions.
        """
        from ..helpers.generator import ListingGenerator
        return ListingGenerator(self.reddit, API_PATH["user_submissions"].format(user=self), *args, **kwargs)

    def __str__(self):
        """
        Returns the Redditor's name.

        Returns
        -------
        name:   str
            The Redditor's username.
        """
        return self.name

    async def moderated_subreddits(self, **kwargs) -> 'Subreddit':
        r"""
        Yields the subreddits the Redditor moderates.

        Parameters
        ----------
        kwargs: \*\*Dict
            ``kwargs`` to be used as query parameters.

        Yields
        ------
        subreddit: Subreddit
            A subreddit the user moderates.
        """
        req = await self.reddit.get_request(API_PATH["moderated"].format(user=self), **kwargs)
        for s in req["data"]:
            yield await self.reddit.subreddit(s["sr"])

    async def message(self, subject, text, from_sr="") -> Dict:
        """
        Message the Redditor.

        Parameters
        ----------
        subject: str
            The subject of the message.
        text: str
            The text contents of the message in markdown.
        from_sr: str
            The subreddit the message is being sent from if applicable.

        Returns
        -------
        resp: Dict
            The response data returned from the endpoint.
        """
        return await self.reddit.message(self.name, subject, text, from_sr)



from typing import TYPE_CHECKING, Dict

from ..helpers.apraw_base import aPRAWBase
from ...const import API_PATH

if TYPE_CHECKING:
    from ..subreddit.subreddit import Subreddit
    from ..reddit.redditor import Redditor
    from ...reddit import Reddit


class SubredditModmail:
    """
    Helper class to aid in retrieving subreddit modmail.

    Members
    -------
    subreddit: Subreddit
        The subreddit this helper operates under.
    """

    def __init__(self, subreddit: 'Subreddit'):
        """
        Create an instance of SubredditModmail.

        Parameters
        ----------
        subreddit: Subreddit
            The subreddit this helper operates under.
        """
        self.subreddit = subreddit

    async def conversations(self) -> 'ModmailConversation':
        """
        Retrieve a list of modmail conversations.

        Yields
        ------
        conversation: ModmailConversation
            A modmail conversation held in the subreddit.
        """
        req = await self.subreddit._reddit.get_request(API_PATH["modmail_conversations"],
                                                       entity=self.subreddit.display_name)
        for id in req["conversations"]:
            yield ModmailConversation(self.subreddit._reddit, req["conversations"][id])


class ModmailConversation(aPRAWBase):
    """
    The model for modmail conversations.

    Members
    -------
    reddit: Reddit
        The :class:`~apraw.Reddit` instance with which requests are made.
    data: Dict
        The data obtained from the /about endpoint.

    **Typical Attributes**

    This table describes attributes that typically belong to objects of this
    class. Attributes are dynamically provided by the :class:`~apraw.models.aPRAWBase` class
    and may vary depending on the status of the response and expected objects.

    ================== ==========================================================================================
    Attribute          Description
    ================== ==========================================================================================
    ``isAuto``         ``bool``
    ``objIds``         A list of dictionaries containing the objects with their IDs and keys.
    ``isRepliable``    Whether the conversation can be replied to.
    ``lastUserUpdate`` A timestamp of the last user update or ``None``.
    ``isInternal``     Whether it's an internal mod conversation.
    ``lastModUpdate``  A timestamp of the last moderator update or ``None``.
    ``lastUpdated``    A timestamp of the last update made overall.
    ``authors``        A list of dictionaries containing authors by name with additional meta information such as
                       ``isMod``, ``isAdmin``, ``isOp``, ``isParticipant``, ``isHidden``, ``id``, ``isDeleted``.
    ``owner``          A dictionary describing the subreddit this conversation is held in.
    ``id``             The ID of this conversation.
    ``isHighlighted``  Whether the conversation has been highlighted.
    ``subject``        The subject of this conversation.
    ``participant``    ``Dict``
    ``state``          ``int``
    ``lastUnread``     ``None``
    ``numMessages``    The number of messages in this conversation.
    ================== ==========================================================================================
    """

    def __init__(self, reddit: 'Reddit', data: Dict,
                 owner: 'Subreddit' = None):
        """
        Create an instance of ``ModmailConversation``.

        Parameters
        ----------
        reddit: Reddit
            The :class:`~apraw.Reddit` instance with which requests are made.
        data: Dict
            The data obtained from the /about endpoint.
        owner: Subreddit
            The subreddit this conversation was held in.
        """
        super().__init__(reddit, data)

        self._data = None
        self._owner = owner

    async def owner(self) -> 'Subreddit':
        """
        Retrieve the owner subreddit of this conversation.

        Returns
        -------
        owner: Subreddit
            The subreddit this conversation was held in.
        """
        if self._owner is None:
            self._owner = await self._reddit.subreddit(self._data["owner"]["displayName"])
        return self._owner

    async def messages(self) -> 'ModmailMessage':
        """
        Retrieve the messages sent in this conversation.

        Yields
        ------
        message: ModmailMessage
            A message sent in this conversation.
        """
        full_data = await self.full_data()
        for msg_id in full_data["messages"]:
            yield ModmailMessage(self, full_data["messages"][msg_id])

    async def full_data(self) -> Dict:
        """
        Retrieve the raw full data from the ``/api/mod/conversations/{id}`` endpoint.

        Returns
        -------
        full_data: Dict
            The full data retrieved from the endpoint.
        """
        if self._data is None:
            self._data = await self._reddit.get_request(API_PATH["modmail_conversation"].format(id=self.id))
        return self._data


class ModmailMessage:
    """
    The model for modmail messages.

    Members
    -------
    conversation: ModmailConversation
        The :class:`~apraw.models.ModmailConversation` instance this message belongs to.
    data: Dict
        The data obtained from the API.
    id: str
        The ID of this message.
    body: str
        The HTML body of this message.
    body_md: str
        The raw body of this message.
    is_internal: str
        Whether the message was sent internally.
    date: str
        A timestamp on which the message was sent.

    .. note::
        ``ModmailMessage`` attributes are loaded statically, meaning they will always be present under the
        abovementioned names.

    """

    def __init__(self, conversation: ModmailConversation, data: Dict):
        """
        Create an instance of a modmail message.

        Parameters
        ----------
        conversation: ModmailConversation
            The :class:`~apraw.models.ModmailConversation` instance this message belongs to.
        data: Dict
            The data obtained from the API.
        """
        self.conversation = conversation
        self.data = data

        self.id = data["id"]

        self.body = data["body"]
        self.body_md = data["bodyMarkdown"]
        self._author = None
        self.is_internal = data["isInternal"]
        self.date = data["date"]

    async def author(self) -> 'Redditor':
        """
        Retrieve the author of this message as a :class:`~apraw.models.Redditor`.

        Returns
        -------
        author: Redditor
            The author of this modmail message.
        """
        if self._author is None:
            if not self.data["author"]["isDeleted"]:
                self._author = self.conversation._reddit.redditor(  # TODO: Add 'await'
                    self.data["author"]["name"])
            else:
                return None
        return self._author



import os
import asyncio
import configparser
import logging
from datetime import datetime, timedelta
from functools import wraps
from typing import Any, Awaitable, Callable, Dict, List, Union

from multidict import CIMultiDictProxy

from .endpoints import API_PATH, BASE_URL
from .models import (Comment, Listing, Redditor, Submission,
                     Subreddit, User, ListingGenerator, Streamable)
from .utils import prepend_kind

if os.path.exists('praw.ini'):
    _prawfile = os.path.abspath('praw.ini')
elif 'APPDATA' in os.environ:  # Windows
    _prawfile = os.path.join(os.environ['APPDATA'], 'praw.ini')
elif 'XDG_CONFIG_HOME' in os.environ:  # Modern Linux
    _prawfile = os.path.join(os.environ['XDG_CONFIG_HOME'], 'praw.ini')
elif 'HOME' in os.environ:  # Legacy Linux
    _prawfile = os.path.join(os.environ['HOME'], '.config', 'praw.ini')


class Reddit:
    """
    The Reddit instance with which root requests can be made.

    Members
    -------
    user: User
        An instance of the logged-in Reddit user.
    """

    def __init__(self, praw_key: str = "", username: str = "", password: str = "",
                 client_id: str = "", client_secret: str = "",
                 user_agent="aPRAW by Dan6erbond"):
        """
        Create a Reddit instance.

        Parameters
        ----------
        praw_key: str
            The key used in a `praw.ini` file instead of manual username, password, client_id and secret.
        username: str
            The Reddit account username.
        password: str
            The Reddit account password.
        client_id: str
            The Reddit script's client_id.
        client_secret
            The Reddit script's client_secret.
        user_agent: str
            User agent to be used in the headers, defaults to "aPRAW by Dan6erbond".
        """
        if praw_key != "":
            config = configparser.ConfigParser()
            config.read(_prawfile)

            self.user = User(self, config[praw_key]["username"], config[praw_key]["password"],
                             config[praw_key]["client_id"], config[praw_key]["client_secret"],
                             config[praw_key]["user_agent"] if "user_agent" in config[praw_key] else user_agent)
        else:
            self.user = User(self, username, password,
                             client_id, client_secret, user_agent)

        self.comment_kind = "t1"
        self.account_kind = "t2"
        self.link_kind = "t3"
        self.message_kind = "t4"
        self.subreddit_kind = "t5"
        self.award_kind = "t6"
        self.modaction_kind = "modaction"
        self.listing_kind = "Listing"
        self.wiki_revision_kind = "WikiRevision"
        self.wikipage_kind = "wikipage"
        self.more_kind = "more"

        self.loop = asyncio.get_event_loop()
        self.request_handler = RequestHandler(self.user)

    @Streamable.streamable
    def subreddits(self, *args, **kwargs):
        r"""
        A :class:`~apraw.models.ListingGenerator` that returns newly created subreddits, which can be streamed using :code:`reddit.subreddits.stream()`.

        Parameters
        ----------
        kwargs: \*\*Dict
            :class:`~apraw.models.ListingGenerator` ``kwargs``.

        Returns
        -------
        generator: ListingGenerator
            A :class:`~apraw.models.ListingGenerator` that retrieves newly created subreddits.
        """
        return ListingGenerator(self, API_PATH["subreddits_new"], *args, **kwargs)

    async def get_request(self, *args, **kwargs):
        """
        Perform an HTTP GET request on the Reddit API.

        Parameters
        ----------
        endpoint: str
            The endpoint to be appended after the base URL (https://oauth.reddit.com/).
        kwargs:
            Query parameters to be appended after the URL.

        Returns
        -------
        resp: Dict or None
            The response JSON data.
        """
        return await self.request_handler.get_request(*args, **kwargs)

    async def post_request(self, *args, **kwargs):
        """
        Perform an HTTP POST request on the Reddit API.

        Parameters
        ----------
        endpoint: str
            The endpoint to be appended after the base URL (https://oauth.reddit.com/).
        url: str
            The direct URL to perform the request on.
        data:
            The data to add to the POST body.
        kwargs:
            Query parameters to be appended after the URL.

        Returns
        -------
        resp: Dict or None
            The response JSON data.
        """
        return await self.request_handler.post_request(*args, **kwargs)

    async def get_listing(self, endpoint: str, subreddit: Subreddit = None, kind_filter: List[str] = None,
                          **kwargs) -> Listing:
        r"""
        Retrieve a listing from an endpoint.

        Parameters
        ----------
        endpoint: str
            The endpoint to be appended after the base URL (https://oauth.reddit.com/).
        subreddit: Subreddit
            The subreddit to dependency inject into retrieved items when possible.
        kind_filter:
            Kinds to return if given, otherwise all are returned.
        kwargs: \*\*Dict
            Query parameters to be appended after the URL.

        Returns
        -------
        listing: Listing
            The listing containing all the endpoint's children.
        """
        resp = await self.get_request(endpoint, **kwargs)
        return Listing(self, resp["data"], kind_filter=kind_filter, subreddit=subreddit)

    async def subreddit(self, display_name: str) -> Subreddit:
        """
        Get a `Subreddit` object according to the given name.

        Parameters
        ----------
        display_name: str
            The display name of the subreddit.

        Returns
        -------
        subreddit: Subreddit
            The subreddit if found.
        result: None
            Returns None if subreddit not found.
        """
        return await Subreddit(self, {"display_name": display_name}).fetch()

    async def info(self, id: str = "", ids: List[str] = [], url: str = ""):
        """
        Get a Reddit item based on its ID or URL.

        Parameters
        ----------
        id: str
            The item's ID.
        ids: List[str]
            Multiple IDs to fetch multiple items at once (max 100).
        url: str
            The item's URL.

        Yields
        -------
        comment: Comment
            A `Comment` object.
        submission: Submission
            A `Submission` object.
        """
        if id:
            for i in await self.get_listing(API_PATH["info"], id=id):
                yield i
        elif ids:
            while ids:
                for i in await self.get_listing(API_PATH["info"], id=",".join(ids[:100])):
                    yield i
                ids = ids[100:]
        elif url:
            for i in await self.get_listing(API_PATH["info"], url=url):
                yield i
        else:
            yield None

    async def submission(self, id: str = "", url: str = "") -> Submission:
        """
        Get a `Submission` object based on its ID or URL.

        Parameters
        ----------
        id: str
            The ID of a submission (with or without kind).
        url: str
            The URL of a submission.

        Returns
        -------
        submission: Submission
            The requested submission.
        """
        if id != "":
            async for link in self.info(prepend_kind(id, self.link_kind)):
                return link
        elif url != "":
            async for link in self.info(url=url):
                return link
        return None

    async def comment(self, id: str = "", url: str = "") -> Comment:
        """
        Get a `Comment` object based on its ID or URL.

        Parameters
        ----------
        id: str
            The ID of a comment (with or without kind).
        url: str
            The URL of a comment.

        Returns
        -------
        comment: Comment
            The requested comment.
        """
        if id != "":
            comment = Comment(self, {"id": id})
            await comment.fetch()
            return comment
        else:
            comment = Comment(self, {"url": url})
            await comment.fetch()
            return comment

    async def close(self):
        await self.request_handler.close()

    async def redditor(self, username: str) -> Redditor:
        """
        Get a `Redditor` object based the Redditor's username.

        Parameters
        ----------
        username: str
            The Redditor's username (without '/u/').

        Returns
        -------
        redditor: Redditor or None
            The requested Redditor, returns None if not found.
        """
        return await Redditor(self, {"username": username}).fetch()

    async def message(self, to: Union[str, Redditor], subject: str, text: str,
                      from_sr: Union[str, Subreddit] = "") -> Dict:
        """
        Message a Redditor or Subreddit.

        Parameters
        ----------
        to: str or Redditor or Subreddit
            The Redditor or Subreddit the message should be sent to.
        subject: str
            The subject of the message.
        text: str
            The text contents of the message.
        from_sr: str or Subreddit
            Optional if the message is being sent from a subreddit.

        Returns
        -------
        result: Dict
            The response JSON data.
        """
        data = {
            "subject": subject,
            "text": text,
            "to": str(to)
        }
        if from_sr != "":
            data["from_sr"] = str(from_sr)
        resp = await self.post_request(API_PATH["compose"], data=data)
        return resp["success"]


class RequestHandler:

    def __init__(self, user: User):
        self.user = user
        self.queue = []

    async def get_request_headers(self) -> Dict:
        if self.user.token_expires <= datetime.now():
            url = "https://www.reddit.com/api/v1/access_token"
            session = await self.user.auth_session()

            headers = {
                "Content-Type": "application/x-www-form-urlencoded",
                "User-Agent": self.user.user_agent
            }

            resp = await session.post(url, data=self.user.password_grant, headers=headers)

            async with resp:
                if resp.status == 200:
                    self.user.access_data = await resp.json()
                    self.user.token_expires = datetime.now(
                    ) + timedelta(seconds=self.user.access_data["expires_in"])
                else:
                    raise Exception("Invalid user data.")

        return {
            "Authorization": "{} {}".format(self.user.access_data["token_type"], self.user.access_data["access_token"]),
            "User-Agent": self.user.user_agent
        }

    def update(self, data: CIMultiDictProxy[str]):
        if "x-ratelimit-remaining" in data:
            self.user.ratelimit_remaining = int(float(data["x-ratelimit-remaining"]))
        if "x-ratelimit-used" in data:
            self.user.ratelimit_used = int(data["x-ratelimit-used"])
        if "x-ratelimit-reset" in data:
            self.user.ratelimit_reset = datetime.now() + timedelta(seconds=int(data["x-ratelimit-reset"]))

    async def close(self):
        await self.user.close()

    class Decorators:

        @classmethod
        def check_ratelimit(
                cls, func: Callable[[Any], Awaitable[Any]]) -> Callable[[Any], Awaitable[Any]]:
            @wraps(func)
            async def execute_request(self, *args, **kwargs) -> Any:
                id = datetime.now().strftime('%Y%m%d%H%M%S')
                self.queue.append(id)

                if self.user.ratelimit_remaining < 1:
                    execution_time = self.user.ratelimit_reset + \
                                     timedelta(seconds=len(self.queue))
                    wait_time = (
                            execution_time -
                            datetime.now()).total_seconds()
                    await asyncio.sleep(wait_time)

                result = await func(self, *args, **kwargs)
                self.queue.remove(id)
                return result

            return execute_request

    @Decorators.check_ratelimit
    async def get_request(self, endpoint: str = "", **kwargs) -> Dict:
        kwargs = {"raw_json": 1, "api_type": "json", **kwargs}
        params = ["{}={}".format(k, kwargs[k]) for k in kwargs]

        url = BASE_URL.format(endpoint, "&".join(params))

        headers = await self.get_request_headers()
        session = await self.user.client_session()
        resp = await session.get(url, headers=headers)

        async with resp:
            self.update(resp.headers)
            return await resp.json()

    @Decorators.check_ratelimit
    async def post_request(self, endpoint: str = "", url: str = "", data: Dict = {}, **kwargs) -> Dict:
        kwargs = {"raw_json": 1, "api_type": "json", **kwargs}
        params = ["{}={}".format(k, kwargs[k]) for k in kwargs]

        if endpoint != "":
            url = BASE_URL.format(endpoint, "&".join(params))
        elif url != "":
            url = "{}?{}".format(url, "&".join(params))

        headers = await self.get_request_headers()
        session = await self.user.client_session()
        resp = await session.post(url, data=data, headers=headers)

        async with resp:
            self.update(resp.headers)
            return await resp.json()



from typing import TYPE_CHECKING, Dict

from ..reddit.comment import Comment
from ..reddit.listing import Listing
from ..reddit.more_comments import MoreComments
from ..subreddit.subreddit import Subreddit

if TYPE_CHECKING:
    from ...reddit import Reddit


class CommentForrest(Listing):

    def __init__(self, reddit: 'Reddit', data: Dict, link_id: str, subreddit: Subreddit = None):
        super().__init__(reddit, data, subreddit, link_id=link_id)
        self._comments_unfolded = []

    async def replace_more(self):
        if not self._comments_unfolded:
            for item in self:
                if isinstance(item, MoreComments):
                    comments = await item.comments()
                    for comment in comments:
                        if comment.replies:
                            await comment.replies.replace_more()
                    self._comments_unfolded.extend(comments)
                elif isinstance(item, Comment):
                    if item.replies:
                        await item.replies.replace_more()
                    self._comments_unfolded.append(item)

            setattr(self, self.CHILD_ATTRIBUTE, self._comments_unfolded)



from enum import Enum
from typing import TYPE_CHECKING, Dict, List, Any, Union

from .redditor import Redditor
from ..helpers.apraw_base import aPRAWBase
from ..helpers.item_moderation import PostModeration
from ..mixins.author import AuthorMixin
from ..mixins.deletable import DeletableMixin
from ..mixins.hideable import HideableMixin
from ..mixins.nsfwable import NSFWableMixin
from ..mixins.replyable import ReplyableMixin
from ..mixins.savable import SavableMixin
from ..mixins.spoilerable import SpoilerableMixin
from ..mixins.subreddit import SubredditMixin
from ..mixins.votable import VotableMixin
from ..subreddit.subreddit import Subreddit
from ...const import API_PATH
from ...utils import prepend_kind

if TYPE_CHECKING:
    from ...reddit import Reddit


class SubmissionKind(Enum):
    """
    An enum representing the valid submission kinds
    """
    LINK = "link"
    SELF = "self"
    IMAGE = "image"
    VIDEO = "video"
    VIDEOGIF = "videogif"


class Submission(aPRAWBase, DeletableMixin, HideableMixin, ReplyableMixin, NSFWableMixin, SavableMixin, VotableMixin,
                 AuthorMixin, SubredditMixin, SpoilerableMixin):
    """
    The model representing submissions.

    Members
    -------
    reddit: Reddit
        The :class:`~apraw.Reddit` instance with which requests are made.
    data: Dict
        The data obtained from the /about endpoint.
    mod: SubmissionModeration
        The :class:`~apraw.models.SubmissionModeration` instance to aid in moderating the submission.
    kind: str
        The item's kind / type.

    **Typical Attributes**

    This table describes attributes that typically belong to objects of this
    class. Attributes are dynamically provided by the :class:`~apraw.models.aPRAWBase` class
    and may vary depending on the status of the response and expected objects.

    ================================= ============================================================================
    Attribute                         Description
    ================================= ============================================================================
    ``all_awardings``                 A list of the awardings on the submission.
    ``allow_live_comments``           Whether live comments have been enabled on this submission.
    ``approved_at_utc``               The UTC timestamp of when the submission was approved.
    ``approved_by``                   The user that approved the submission.
    ``approved``                      Whether the submission has been approved by the moderators of the subreddit.
    ``archived``                      Whether the submission has been archived by Reddit.
    ``author_flair_background_color`` The submission author's flair background color.
    ``author_flair_css_class``        The submission's author flair CSS class.
    ``author_flair_richtext``         The submission's author flair text.
    ``author_flair_template_id``      The submission author's flair template ID if applicable.
    ``author_flair_text_color``       The submission's author flair text color if applicable.
    ``author_flair_text``             The author's flair text if applicable.
    ``author_flair_type``             The type of flair used by the submission's author.
    ``author_fullname``               The author of the submission prepended with ``t2_``.
    ``author_patreon_flair``          The submission's author Patreon flair.
    ``author``                        The name of the submission's Redditor.
    ``banned_at_utc``                 The UTC timestamp at which the author was banned.
    ``banned_by``                     ``null``
    ``can_gild``                      Whether the logged-in user can gild the submission.
    ``can_mod_post``                  Whether the logged-in user can modify the post.
    ``category``                      The submission's category.
    ``clicked``                       Whether the submission has been clicked by the logged-in user previously.
    ``content_categories``            The content categories assigned to the submission.
    ``contest_mode``                  Whether the moderators of the subreddit have enabled contest mode on
                                      the submission.
    ``created_utc``                   The parsed UTC ``datetime`` on which the submission was made.
    ``created``                       The timestamp of when the submission was posted.
    ``discussion_type``               ``null``
    ``distinguished``                 The type of distinguishment on the submission.
    ``domain``                        The domain of the submission.
    ``downs``                         The number of downvotes on the submission.
    ``edited``                        Whether the submission has been edited by its author.
    ``gilded``                        The number of awards this submission has received.
    ``gildings``                      The gild awards the submission has received.
    ``hidden``                        Whether the submission has been hidden by the logged-in user.
    ``hide_score``                    Whether clients should hide the score from users.
    ``id``                            The submission's ID.
    ``ignore_reports``                Whether reports should be ignored on this submission.``
    ``is_crosspostable``              Whether the submission can be crossposted to other subreddits.
    ``is_meta``                       Whether the submission is a meta post.
    ``is_original_content``           Whether the submission has been marked as original content.
    ``is_reddit_media_domain``        Whether the media has been uploaded to Reddit.
    ``is_robot_indexable``            Whether the submission can be indexed by robots.
    ``is_self``                       Whether the submission is a self post.
    ``is_video``                      Whether the submission is a video post.
    ``likes``                         ``bool``
    ``link_flair_background_color``   The submission's flair background color.
    ``link_flair_css_class``          The CSS class applied on the submission's flair if applicable.
    ``link_flair_richtext``           The submission's flair text if applicable.
    ``link_flair_template_id``        The submission's flair template ID if applicable.
    ``link_flair_text_color``         The submission's flair text color if applicable.
    ``link_flair_text``               The submission's flair text.
    ``link_flair_type``               The type of flair applied to the submission.
    ``locked``                        Whether the submission has been locked by the subreddit moderators.
    ``media_embed``                   ``Dict``
    ``media_only``                    Whether the submission only consists of media.
    ``media``                         ``null``
    ``mod_note``                      Moderator notes added to the submission.
    ``mod_reason_by``                 The moderator who added the removal reason if applicable.
    ``mod_reason_title``              The reason the submission has been removed by moderators if applicable.
    ``mod_reports``                   A list of moderator reports on the submission.
    ``name``                          The ID of the submission prepended with ``t3_``.
    ``no_follow``                     ``bool``
    ``num_comments``                  The number of comments on the submission.
    ``num_crossposts``                The number of times the submission has been crossposted.
    ``num_reports``                   The number of reports on the submission.
    ``over_18``                       Whether the submission has been marked as NSFW.
    ``parent_whitelist_status``       ``null``
    ``permalink``                     The submission's permalink.
    ``pinned``                        Whether the submission has been pinned on the subreddit.
    ``pwls``                          ``null``
    ``quarantine``                    Whether the submission was posted in a quarantined subreddit.
    ``removal_reason``                The submission's removal reason if applicable.
    ``removed``                       Whether the submission has been removed by the subreddit moderators.
    ``report_reasons``                A list of report reasons on the submission.
    ``saved``                         Whether the submission has been saved by the logged-in user.
    ``score``                         The overall submission vote score.
    ``secure_media_embed``            ``Dict``
    ``secure_media``                  ``null``
    ``selftext_html``                 The submission text as HTML.
    ``selftext``                      The submission's selftext.
    ``send_replies``                  Whether the author of the submission will receive reply notifications.
    ``spam``                          Whether the submission has been marked as spam.
    ``spoiler``                       Whether the submission contains a spoiler.
    ``stickied``                      Whether the submission is stickied on the subreddit.
    ``subreddit_id``                  The subreddit's ID prepended with ``t5_``.
    ``subreddit_name_prefixed``       The name of the subreddit the submission was posted on, prefixed with "r/".
    ``subreddit_subscribers``         The number of subscribers to the submission's subreddit.
    ``subreddit_type``                The type of the subreddit the submission was posted on
                                      (public, restricted, private).
    ``subreddit``                     The name of the subreddit on which the submission was posted.
    ``suggested_sort``                The suggested sort method for comments.
    ``thumbnail_height``              The height of the submission's thumbnail if applicable.
    ``thumbnail_width``               The width of the submission's thumbnail if applicable.
    ``thumbnail``                     A URL to the submission's thumbnail if applicable.
    ``title``                         The submission's title.
    ``total_awards_received``         The number of awards on the submission.
    ``ups``                           The number of upvotes on the submission.
    ``url``                           The full URL of the submission.
    ``user_reports``                  A list of the user reports on the submission.
    ``view_count``                    The number of views on the submission.
    ``visited``                       Whether the logged-in user has visited the submission previously.
    ``whitelist_status``              ``null``
    ``wls``                           ``null``
    ================================= ============================================================================

    .. note::
        Many of these attributes are only available if the logged-in user has moderator access to the item.

    """

    def __init__(self, reddit: 'Reddit', data: Dict, subreddit: Subreddit = None, author: Redditor = None):
        """
        Create an instance of a submission object.

        Parameters
        ----------
        reddit: Reddit
            The :class:`~apraw.Reddit` instance with which requests are made.
        data: Dict
            The data obtained from the /about endpoint.
        full_data: Dict
            The full_data retrieved by the /r/{sub}/comments/{id} endpoint.
        subreddit: Subreddit
            The subreddit this submission was posted on.
        author: Redditor
            The author of this submission as a :class:`~apraw.models.Redditor`.
        """
        self.comments = list()

        aPRAWBase.__init__(self, reddit, data, reddit.link_kind)
        AuthorMixin.__init__(self, author)
        SubredditMixin.__init__(self, subreddit)

        self.mod = SubmissionModeration(reddit, self)

    async def fetch(self):
        """
        Fetch this item's information from a suitable API endpoint.

        Returns
        -------
        self: Submission
            The updated model.
        """
        if "subreddit" in self._data and "id" in self._data:
            resp = await self._reddit.get_request(
                API_PATH["submission"].format(sub=self._data["subreddit"], id=self._data["id"]))
            self._update(resp)
        elif "id" in self._data:
            resp = await self._reddit.get_request(API_PATH["info"],
                                                  id=prepend_kind(self._data["id"], self._reddit.link_kind))
            self._update(resp["data"]["children"][0]["data"])
        return self

    def _update(self, _data: Union[List, Dict[str, Any]]):
        """
        Update the base with new information.

        Parameters
        ----------
        _data: Dict
            The data obtained from the API.
        """
        if isinstance(_data, dict) or isinstance(_data, list):
            if isinstance(_data, dict):
                data = _data
            else:
                from ..helpers.comment_forrest import CommentForrest
                self.comments = CommentForrest(self._reddit, _data[1]["data"], self.fullname)
                data = _data[0]["data"]["children"][0]

            data["original_content"] = data.get("is_original_content", False)
            super()._update(data)
        else:
            raise ValueError("data is not of type 'dict' or 'list'.")


class SubmissionModeration(PostModeration, NSFWableMixin, SpoilerableMixin):
    """
    A helper class to moderate submissions.

    Members
    -------
    reddit: Reddit
        The :class:`~apraw.Reddit` instance with which requests are made.
    fullname: str
        The ID prepended with the kind of the item this helper belongs to.
    """

    def __init__(self, reddit: 'Reddit', submission: Submission):
        """
        Create an instance of ``SubmissionModeration``.

        Parameters
        ----------
        reddit : Reddit
            The :class:`~apraw.Reddit` instance with which requests are made.
        submission : Submission
            The submission this helper performs requests for.
        """
        super().__init__(reddit, submission)

    async def sticky(self, position: int = 1, to_profile: bool = False):
        """
        Sticky a submission in its subreddit.

        Parameters
        ----------
        position : int
            The "slot" the submission will be stickied to.
        to_profile : bool
            Whether the submission will be stickied to the user profile.

        Returns
        -------
        resp: Dict
            The API response JSON.
        """
        return await self.reddit.post_request(API_PATH["mod_sticky"], **{
            "id": self.fullname,
            "num": position,
            "state": True,
            "to_profile": to_profile
        })

    async def unsticky(self, to_profile: bool = False):
        """
        Unsticky a submission from its subreddit.

        Parameters
        ----------
        to_profile : bool
            Whether the submission will be unstickied from the user profile.

        Returns
        -------
        resp: Dict
            The API response JSON.
        """
        return await self.reddit.post_request(API_PATH["mod_sticky"], **{
            "id": self.fullname,
            "state": False,
            "to_profile": to_profile
        })



from typing import TYPE_CHECKING, AsyncIterator, Dict, Union, Any

from .modmail import SubredditModmail
from .moderation import SubredditModerator, SubredditModeration
from .wiki import SubredditWiki
from ..helpers.apraw_base import aPRAWBase
from ..helpers.streamable import Streamable
from ...const import API_PATH

if TYPE_CHECKING:
    from ..reddit.submission import SubmissionKind, Submission
    from ...reddit import Reddit


class Subreddit(aPRAWBase):
    """
    The model representing subreddits.

    Members
    -------
    kind: str
        The item's kind / type.
    mod: SubredditModeration
        Returns an instance of :class:`~apraw.models.SubredditModeration`.
    modmail: SubredditModmail
        Returns an instance of :class:`~apraw.models.SubredditModmail`.
    wiki: SubredditWiki
        Returns an instance of :class:`~apraw.models.SubredditWiki`.

    **Examples**

    To grab new submissions made on a subreddit:

    .. code-block:: python3

        sub = await reddit.subreddit("aprawtest")
        async for submission in sub.new(): # use .new.stream() for endless polling
            print(submission.title, submission.body)

    **Typical Attributes**

    This table describes attributes that typically belong to objects of this
    class. Attributes are dynamically provided by the :class:`~apraw.models.aPRAWBase` class
    and may vary depending on the status of the response and expected objects.

    ================================ ==================================================================
    Attribute                        Description
    ================================ ==================================================================
    ``accounts_active_is_fuzzed``    ``bool``
    ``accounts_active``              ``null``
    ``active_user_count``            The number of active users on the subreddit.
    ``advertiser_category``          ``string``
    ``all_original_content``         Whether the subreddit requires all content to be OC.
    ``allow_discovery``              Whether the subreddit can be discovered.
    ``allow_images``                 Whether images are allowed as submissions.
    ``allow_videogifs``              Whether GIFs are allowed as submissions.
    ``allow_videos``                 Whether videos are allowed as submissions.
    ``banner_background_color``      The banner's background color if applicable, otherwise empty.
    ``banner_background_image``      A URL to the subreddit's banner image.
    ``banner_img``                   A URL to the subreddit's banner image if applicable.
    ``banner_size``                  The subreddit's banner size if applicable.
    ``can_assign_link_flair``        Whether submission flairs can be assigned.
    ``can_assign_user_flair``        Whether the user can assign their own flair on the subreddit.
    ``collapse_deleted_comments``    Whether deleted comments should be deleted by clients.
    ``comment_score_hide_mins``      The minimum comment score to hide.
    ``community_icon``               A URL to the subreddit's community icon if applicable.
    ``created_utc``                  The date on which the subreddit was created in UTC ``datetime``.
    ``created``                      The time the subreddit was created on.
    ``description_html``             The subreddit's description as HTML.
    ``description``                  The subreddit's short description.
    ``disable_contributor_requests`` ``bool``
    ``display_name_prefixed``        The subreddit's display name prefixed with 'r/'.
    ``display_name``                 The subreddit's display name.
    ``emojis_custom_size``           The custom size set for emojis.
    ``emojis_enabled``               Whether emojis are enabled on this subreddit.
    ``free_form_reports``            Whether it's possible to submit free form reports.
    ``has_menu_widget``              Whether the subreddit has menu widgets.
    ``header_img``                   A URL to the subreddit's header image of applicable.
    ``header_size``                  The subreddit's header size.
    ``header_title``                 The subreddit's header title.
    ``hide_ads``                     Whether ads are hidden on this subreddit.
    ``icon_img``                     A URL to the subreddit's icon image of applicable.
    ``icon_size``                    The subreddit's icon size.
    ``id``                           The subreddit's ID.
    ``is_enroled_in_new_modmail``    Whether the subreddit is enrolled in new modmail.
    ``key_color``                    ``string``
    ``lang``                         The subreddit's language.
    ``link_flair_enabled``           Whether link flairs have been enabled for the subreddit.
    ``link_flair_position``          The position of link flairs.
    ``mobile_banner_size``           A URL to the subreddit's mobile banner if applicable.
    ``name``                         The subreddit's fullname (t5_ID).
    ``notification_level``
    ``original_content_tag_enabled`` Whether the subreddit has the OC tag enabled.
    ``over18``                       Whether the subreddit is NSFW.
    ``primary_color``                The subreddit's primary color.
    ``public_description_html``      The subreddit's public description as HTML.
    ``public_description``           The subreddit's public description string.
    ``public_traffic``               ``bool``
    ``quarantine``                   Whether the subreddit is quarantined.
    ``restrict_commenting``             Whether comments by users are restricted on the subreddit.
    ``restrict_posting``             Whether posts to the subreddit are restricted.
    ``show_media_preview``           Whether media previews should be displayed by clients.
    ``show_media``
    ``spoilers_enabled``             Whether the spoiler tag is enabled on the subreddit.
    ``submission_type``              The types of allowed submissions. Default is "any".
    ``submit_link_label``            The subreddit's submit label if applicable.
    ``submit_text_html``             The HTML submit text if a custom one is set on the subreddit.
    ``submit_text_label``            The text used for the submit button.
    ``submit_text``                  The markdown submit text if a custom one is set on the subreddit.
    ``subreddit_type``               The subreddit type, either "public", "restricted" or "private".
    ``subscribers``                  The number of subreddit subscribers.
    ``suggested_comment_sort``       The suggested comment sort algorithm, can be ``null``.
    ``title``                        The subreddit's banner title.
    ``url``                          The subreddit's display name prepended with "/r/".
    ``user_can_flair_in_sr``         Whether the user can assign custom flairs (nullable).
    ``user_flair_background_color``  The logged in user's flair background color if applicable.
    ``user_flair_css_class``         The logged in user's flair CSS class.
    ``user_flair_enabled_in_sr``     Whether the logged in user's subreddit flair is enabled.
    ``user_flair_position``          The position of user flairs on the subreddit (right or left).
    ``user_flair_richtext``          The logged in user's flair text if applicable.
    ``user_flair_template_id``       The logged in user's flair template ID if applicable.
    ``user_flair_text_color``        The logged in user's flair text color.
    ``user_flair_text``              The logged in user's flair text.
    ``user_flair_type``              The logged in user's flair type.
    ``user_has_favorited``           Whether the logged in user has favorited the subreddit.
    ``user_is_banned``               Whether the logged in user is banned from the subreddit.
    ``user_is_contributor``          Whether the logged in user has contributed to the subreddit.
    ``user_is_moderator``            Whether the logged in user is a moderator on the subreddit.
    ``user_is_muted``                Whether the logged in user has been muted by the subreddit.
    ``user_is_subscriber``           Whether the logged in user is subscribed to the subreddit.
    ``user_sr_flair_enabled``        Whether the logged in user's subreddit flair is enabled.
    ``user_sr_theme_enabled``        Whether the logged in user has enabled the custom subreddit theme.
    ``videostream_links_count``      The number of submissions with videostream links.
    ``whitelist_status``
    ``wiki_enabled``                 Whether the subreddit has the wiki enabled.
    ``wls``                          ``null``
    ================================ ==================================================================
    """

    def __init__(self, reddit: 'Reddit', data: Dict = None):
        """
        Create a Subreddit instance.

        Parameters
        ----------
        reddit: Reddit
            The :class:`~apraw.Reddit` instance with which requests are made.
        data: Dict
            The data obtained from the /about endpoint.
        """
        super().__init__(reddit, data, reddit.subreddit_kind)

        self.mod = SubredditModeration(self)
        self.modmail = SubredditModmail(self)
        self.wiki = SubredditWiki(self)

    async def fetch(self):
        """
        Fetch this item's information from a suitable API endpoint.

        Returns
        -------
        self: Subreddit
            The ``Subreddit`` model with updated data.
        """
        resp = await self._reddit.get_request(API_PATH["subreddit_about"].format(sub=self.display_name))
        self._update(resp["data"])
        return self

    def _update(self, data: Dict[str, Any]):
        """
        Update the base with new information.

        Parameters
        ----------
        data: Dict
            The data obtained from the /about endpoint.
        """
        super()._update(data)
        self.quarantine = data["quarantine"] if "quarantine" in data else False

    async def random(self):
        """
        Retrieve a random submission from the subreddit.

        Returns
        -------
        submission: Submission
            A random submission from the subreddit.
        """
        resp = await self._reddit.get_request(API_PATH["subreddit_random"].format(sub=self))
        from ..reddit.listing import Listing
        listing = Listing(self._reddit, data=resp[0]["data"], subreddit=self)
        return next(listing)

    @Streamable.streamable
    def comments(self, *args, **kwargs):
        r"""
        Returns an instance of :class:`~apraw.models.ListingGenerator` mapped to the comments endpoint.

        .. note::
            This listing can be streamed doing the following:

            .. code-block:: python3

                for comment in subreddit.comments.stream():
                    print(comment)

        Parameters
        ----------
        kwargs: \*\*Dict
            :class:`~apraw.models.ListingGenerator` ``kwargs``.

        Returns
        -------
        generator: ListingGenerator
            A :class:`~apraw.models.ListingGenerator` mapped to the comments endpoint.
        """
        from ..helpers.generator import ListingGenerator
        return ListingGenerator(self._reddit, API_PATH["subreddit_comments"].format(sub=self.display_name),
                                subreddit=self, *args, **kwargs)

    @Streamable.streamable
    def new(self, *args, **kwargs):
        r"""
        Returns an instance of :class:`~apraw.models.ListingGenerator` mapped to the new submissions endpoint.

        .. note::
            This listing can be streamed doing the following:

            .. code-block:: python3

                for comment in submissions.new.stream():
                    print(comment)

        Parameters
        ----------
        kwargs: \*\*Dict
            :class:`~apraw.models.ListingGenerator` ``kwargs``.

        Returns
        -------
        generator: ListingGenerator
            A :class:`~apraw.models.ListingGenerator` mapped to the new submissions endpoint.
        """
        from ..helpers.generator import ListingGenerator
        return ListingGenerator(self._reddit, API_PATH["subreddit_new"].format(sub=self.display_name), subreddit=self,
                                *args, **kwargs)

    def hot(self, *args, **kwargs):
        r"""
        Returns an instance of :class:`~apraw.models.ListingGenerator` mapped to the hot submissions endpoint.

        Parameters
        ----------
        kwargs: \*\*Dict
            :class:`~apraw.models.ListingGenerator` ``kwargs``.

        Returns
        -------
        generator: ListingGenerator
            A :class:`~apraw.models.ListingGenerator` mapped to the hot submissions endpoint.
        """
        from ..helpers.generator import ListingGenerator
        return ListingGenerator(self._reddit, API_PATH["subreddit_hot"].format(sub=self.display_name), subreddit=self,
                                *args, **kwargs)

    def rising(self, *args, **kwargs):
        r"""
        Returns an instance of :class:`~apraw.models.ListingGenerator` mapped to the rising submissions endpoint.

        Parameters
        ----------
        kwargs: \*\*Dict
            :class:`~apraw.models.ListingGenerator` ``kwargs``.

        Returns
        -------
        generator: ListingGenerator
            A :class:`~apraw.models.ListingGenerator` mapped to the rising submissions endpoint.
        """
        from ..helpers.generator import ListingGenerator
        return ListingGenerator(self._reddit, API_PATH["subreddit_rising"].format(sub=self.display_name), *args,
                                **kwargs)

    def top(self, *args, **kwargs):
        r"""
        Returns an instance of :class:`~apraw.models.ListingGenerator` mapped to the top submissions endpoint.

        Parameters
        ----------
        kwargs: \*\*Dict
            :class:`~apraw.models.ListingGenerator` ``kwargs``.

        Returns
        -------
        generator: ListingGenerator
            A :class:`~apraw.models.ListingGenerator` mapped to the top submissions endpoint.
        """
        from ..helpers.generator import ListingGenerator
        return ListingGenerator(self._reddit, API_PATH["subreddit_top"].format(sub=self.display_name), subreddit=self,
                                *args, **kwargs)

    def __str__(self):
        """
        Returns the subreddit's name as a string.

        Returns
        -------
        display_name: str
            The subreddit's name as a string.
        """
        return self.display_name

    async def moderators(self, **kwargs) -> AsyncIterator['SubredditModerator']:
        r"""
        Yields all the subreddit's moderators.

        Parameters
        ----------
        kwargs: \*\*Dict
            The query parameters to be added to the GET request.

        Yields
        ------
        moderator: SubredditModerator
            An instance of the moderators as :class:`~apraw.models.SubredditModerator`.
        """
        req = await self._reddit.get_request(API_PATH["subreddit_moderators"].format(sub=self.display_name), **kwargs)
        for u in req["data"]["children"]:
            yield SubredditModerator(self._reddit, u)

    async def message(self, subject: str, text: str, from_sr: Union[str, 'Subreddit'] = "") -> Dict:
        """
        Send a message to the subreddit.

        Parameters
        ----------
        subject: str
            The message subject.
        text: str
            The message contents as markdown.
        from_sr: str or Subreddit
            The subreddit the message is being sent from if applicable.
        Returns
        -------
        response: Dict
            The API response JSON as a dictionary.
        """
        return await self._reddit.message(API_PATH["subreddit"].format(sub=self.display_name), subject, text,
                                          str(from_sr))

    async def submit(self, title: str, kind: 'SubmissionKind', **kwargs) -> 'Submission':
        """
        Make a new post to the subreddit.
        If `kind` is SubmissionKind.LINK then `url` is expected to be a valid url,
        otherwise `text` is expected (and it can be markdown text)

        Parameters
        -------
        title: str
            The post's title.
        kind: SubmissionKind
            The post's kind.
        url: str
            Optional, the url if kind is LINK.
        text: str
            Optional, the text body of the post.
        nsfw: bool = False
            If the post if nsfw or not.
        resubmit: bool = False
            If the post is a re-submit or not.
            Needs to be True if a link with the same URL has already been submitted to the specified subreddi
        spoiler: bool = False
            If the post is a spoiler or not.
        """
        from ..reddit.submission import Submission, SubmissionKind

        url = kwargs.get("url", None)
        text = kwargs.get("text", None)

        if kind == SubmissionKind.LINK and not url:
            raise ValueError("A url was expected")
        if kind == SubmissionKind.SELF and not text:
            raise ValueError("A text body was expected")

        resp = await self._reddit.post_request(API_PATH["submit"], **{
            "sr": str(self),
            "title": title,
            "kind": kind.value,
            "url": url,
            "text": text,
            "nsfw": kwargs.get("nsfw", False),
            "resubmit": kwargs.get("resubmit", False),
            "spoiler": kwargs.get("spoiler", False)
        })

        submission = Submission(self._reddit, {"id": resp["json"]["data"]["id"]})
        await submission.fetch()

        return submission



from datetime import datetime
from typing import Dict, TYPE_CHECKING

from ..helpers.apraw_base import aPRAWBase
from ..helpers.streamable import Streamable
from ..reddit.redditor import Redditor
from ...const import API_PATH

if TYPE_CHECKING:
    from ...reddit import Reddit
    from .subreddit import Subreddit


class SubredditModerator(aPRAWBase):
    """
    The model representing subreddit moderators. Redditors can be retrieved via ``redditor()``.

    **Typical Attributes**

    This table describes attributes that typically belong to objects of this
    class. Attributes are dynamically provided by the :class:`~apraw.models.aPRAWBase` class
    and may vary depending on the status of the response and expected objects.

    ========================== ============================================================
    Attribute                  Description
    ========================== ============================================================
    ``added``                  The date on which the moderator was added.
    ``author_flair_css_class`` The moderator's flair CSS class in the respective subreddit.
    ``author_flair_text``      The moderator's flair text in the respective subreddit.
    ``id``                     The Redditor's fullname (t2_ID).
    ``mod_permissions``        A list of all the moderator permissions or ``["all"]``.
    ``name``                   The Redditor's name.
    ========================== ============================================================
    """

    def __init__(self, reddit: 'Reddit', data: Dict):
        """
        Create a SubredditModerator instance.

        Parameters
        ----------
        reddit: Reddit
            The :class:`~apraw.Reddit` instance with which requests are made.
        data: Dict
            The data obtained from the API.
        """
        super().__init__(reddit, data)

        self.added = datetime.utcfromtimestamp(data["date"])

    def __str__(self):
        """
        Returns the Redditor's name.

        Returns
        -------
        name: str
            The Redditor's name.
        """
        return self.name

    async def redditor(self) -> Redditor:
        """
        Retrieve the Redditor this Moderator represents.

        Returns
        -------
        redditor: Redditor
            The Redditor that is represented by this object.
        """
        return await self._reddit.redditor(self.name)


class SubredditModeration:
    """
    A helper class for grabbing listings to Subreddit moderation items.

    Members
    -------
    subreddit: Subreddit
        The subreddit this helper instance belongs to and performs requests for.
    """

    def __init__(self, subreddit: 'Subreddit'):
        """
        Create an instance of ``SubredditModeration``.

        Parameters
        ----------
        subreddit: Subreddit
            The subreddit this helper instance belongs to and performs requests for.
        """
        self.subreddit = subreddit

    @Streamable.streamable
    def reports(self, *args, **kwargs):
        r"""
        Returns an instance of :class:`~apraw.models.ListingGenerator` mapped to grab reported items.

        .. note::
            This listing can be streamed doing the following:

            .. code-block:: python3

                for comment in subreddit.mod.reports.stream():
                    print(comment)

        Parameters
        ----------
        kwargs: \*\*Dict
            :class:`~apraw.models.ListingGenerator` ``kwargs``.

        Returns
        -------
        generator: ListingGenerator
            A :class:`~apraw.models.ListingGenerator` mapped to grab reported items.
        """
        from ..helpers.generator import ListingGenerator
        return ListingGenerator(self.subreddit._reddit,
                                API_PATH["subreddit_reports"].format(sub=self.subreddit.display_name),
                                subreddit=self.subreddit, *args, **kwargs)

    @Streamable.streamable
    def spam(self, *args, **kwargs):
        r"""
        Returns an instance of :class:`~apraw.models.ListingGenerator` mapped to grab items marked as spam.

        .. note::
            This listing can be streamed doing the following:

            .. code-block:: python3

                for comment in subreddit.mod.spam.stream():
                    print(comment)

        Parameters
        ----------
        kwargs: \*\*Dict
            :class:`~apraw.models.ListingGenerator` ``kwargs``.

        Returns
        -------
        generator: ListingGenerator
            A :class:`~apraw.models.ListingGenerator` mapped to grab items marked as spam.
        """
        from ..helpers.generator import ListingGenerator
        return ListingGenerator(self.subreddit._reddit,
                                API_PATH["subreddit_spam"].format(sub=self.subreddit.display_name),
                                subreddit=self.subreddit, *args, **kwargs)

    @Streamable.streamable
    def modqueue(self, *args, **kwargs):
        r"""
        Returns an instance of :class:`~apraw.models.ListingGenerator` mapped to grab items in the modqueue.

        .. note::
            This listing can be streamed doing the following:

            .. code-block:: python3

                for comment in subreddit.mod.modqueue.stream():
                    print(comment)

        Parameters
        ----------
        kwargs: \*\*Dict
            :class:`~apraw.models.ListingGenerator` ``kwargs``.

        Returns
        -------
        generator: ListingGenerator
            A :class:`~apraw.models.ListingGenerator` mapped to grab items in the modqueue.
        """
        from ..helpers.generator import ListingGenerator
        return ListingGenerator(self.subreddit._reddit,
                                API_PATH["subreddit_modqueue"].format(sub=self.subreddit.display_name),
                                subreddit=self.subreddit, *args, **kwargs)

    @Streamable.streamable
    def unmoderated(self, *args, **kwargs):
        r"""
        Returns an instance of :class:`~apraw.models.ListingGenerator` mapped to grab unmoderated items.

        .. note::
            This listing can be streamed doing the following:

            .. code-block:: python3

                for comment in subreddit.mod.unmoderated.stream():
                    print(comment)

        Parameters
        ----------
        kwargs: \*\*Dict
            :class:`~apraw.models.ListingGenerator` ``kwargs``.

        Returns
        -------
        generator: ListingGenerator
            A :class:`~apraw.models.ListingGenerator` mapped to grab unmoderated items.
        """
        from ..helpers.generator import ListingGenerator
        return ListingGenerator(self.subreddit._reddit,
                                API_PATH["subreddit_unmoderated"].format(sub=self.subreddit.display_name),
                                subreddit=self.subreddit, *args, **kwargs)

    @Streamable.streamable
    def edited(self, *args, **kwargs):
        r"""
        Returns an instance of :class:`~apraw.models.ListingGenerator` mapped to grab edited items.

        .. note::
            This listing can be streamed doing the following:

            .. code-block:: python3

                for comment in subreddit.mod.edited.stream():
                    print(comment)

        Parameters
        ----------
        kwargs: \*\*Dict
            :class:`~apraw.models.ListingGenerator` ``kwargs``.

        Returns
        -------
        generator: ListingGenerator
            A :class:`~apraw.models.ListingGenerator` mapped to grab edited items.
        """
        from ..helpers.generator import ListingGenerator
        return ListingGenerator(self.subreddit._reddit,
                                API_PATH["subreddit_edited"].format(sub=self.subreddit.display_name),
                                subreddit=self.subreddit, *args, **kwargs)

    @Streamable.streamable
    def log(self, *args, **kwargs):
        r"""
        Returns an instance of :class:`~apraw.models.ListingGenerator` mapped to grab mod actions in the subreddit log.

        .. note::
            This listing can be streamed doing the following:

            .. code-block:: python3

                for comment in subreddit.mod.log.stream():
                    print(comment)

        Parameters
        ----------
        kwargs: \*\*Dict
            :class:`~apraw.models.ListingGenerator` ``kwargs``.

        Returns
        -------
        generator: ListingGenerator
            A :class:`~apraw.models.ListingGenerator` mapped to grab mod actions in the subreddit log.
        """
        from ..helpers.generator import ListingGenerator
        return ListingGenerator(self.subreddit._reddit,
                                API_PATH["subreddit_log"].format(sub=self.subreddit.display_name),
                                subreddit=self.subreddit, *args, **kwargs)


class ModAction(aPRAWBase):
    """
    A model representing mod actions taken on specific items.

    Members
    -------
    reddit: Reddit
        The :class:`~apraw.Reddit` instance with which requests are made.
    data: Dict
        The data obtained from the /about endpoint.
    kind: str
        The item's kind / type.

    **Typical Attributes**

    This table describes attributes that typically belong to objects of this
    class. Attributes are dynamically provided by the :class:`~apraw.models.aPRAWBase` class
    and may vary depending on the status of the response and expected objects.

    =========================== =========================================================================
    Attribute                   Description
    =========================== =========================================================================
    ``action``                  The type of action performed.
    ``created_utc``             The parsed UTC datetime of when the action was performed.
    ``description``             The description added to the action if applicable.
    ``details``                 The details of the action performed.
    ``id``                      The ID of the mod action prepended with "ModAction_".
    ``mod_id36``                The ID36 of the moderator who performed the action.
    ``mod``                     The username of the moderator who performed the action.
    ``sr_id36``                 The ID36 of the subreddit the action was performed on.
    ``subreddit_name_prefixed`` The name of the subreddit the action was performed on prefixed with "r/".
    ``subreddit``               The name of the subreddit the action was performed on.
    ``target_author``           The author of the target item if applicable.
    ``target_body``             The body of the target item if applicable.
    ``target_fullname``         The id of the target with its kind prepended. (e.g. "t3_d5229o")
    ``target_permalink``        The target of the comment or submission if applicable.
    ``target_title``            The title of the submission if applicable.
    =========================== =========================================================================
    """

    def __init__(self, reddit, data, subreddit=None):
        """
        Create an instance of a ModAction.

        Parameters
        ----------
        data: Dict
            The data returned from the API endpoint.
        subreddit: Subreddit
            The subreddit this ModAction belongs to.
        """
        super().__init__(reddit, data, reddit.modaction_kind)

        self.subreddit = subreddit

    async def mod(self) -> Redditor:
        """
        Returns the Redditor who performed this action.

        Returns
        -------
        redditor: Redditor
            The Redditor who performed this action.
        """
        return await self.subreddit._reddit.redditor(self.mod)



from ..reddit.redditor import Redditor


class AuthorMixin:
    """
    Mixin for items with an author.
    """

    def __init__(self, author: Redditor = None):
        self._author = author

    async def author(self) -> Redditor:
        """
        Retrieve the item's author as a :class:`~apraw.models.Redditor`.

        Returns
        -------
        author: Redditor
            The item's author.
        """
        if self._author is None:
            self._author = await self._reddit.redditor(self._data["author"])
        return self._author



from ...const import API_PATH


class SavableMixin:
    """
    Mixin for savable objects.
    """

    async def save(self, category: str = ""):
        """
        Save the item in a category.

        Parameters
        ----------
        category : str, optional
            The category name.

        Returns
        -------
        resp: Dict
            The API response JSON.
        """
        return await self.reddit.post_request(API_PATH["post_save"], id=self.fullname, category=category)

    async def unsave(self):
        """
        Unsave the item.

        Returns
        -------
        resp: Dict
            The API response JSON.
        """
        return await self.reddit.post_request(API_PATH["post_unsave"], id=self.fullname)



from ..subreddit.subreddit import Subreddit


class SubredditMixin:
    """
    Mixin for items in a subreddit.
    """

    def __init__(self, subreddit: Subreddit = None):
        self._subreddit = subreddit

    async def subreddit(self) -> Subreddit:
        """
        Retrieve the subreddit this item was made in as a :class:`~apraw.models.Subreddit`.

        Returns
        -------
        subreddit: Subreddit
            The subreddit this item was made in.
        """
        if self._subreddit is None:
            self._subreddit = await self._reddit.subreddit(self._data["subreddit"])
        return self._subreddit



import random


class ExponentialCounter:
    """A class to provide an exponential counter with jitter."""

    def __init__(self, max_counter: int):
        """
        Initialize an instance of ExponentialCounter.

        Parameters
        ----------
        max_counter: int
            The maximum value to reach before resetting.
        """
        self._value = 1
        self._max = max_counter

    def count(self) -> int:
        """
        Increment the counter and return the current value with jitter.

        Returns
        -------
        value: int
            The current value defined by the counter.
        """
        max_jitter = self._value / 16.0
        self._value = min(self._value * 2, self._max)
        value = self._value + random.random() * max_jitter - max_jitter / 2
        return value

    def reset(self):
        """
        Reset the counter to 1.

        Returns
        -------
        value: int
            The reset value for easy use.
        """
        self._value = 1
        return self._value



from typing import TYPE_CHECKING, Dict, Iterator, List

from .comment import Comment
from .message import Message
from .more_comments import MoreComments
from .submission import Submission
from ..helpers.apraw_base import aPRAWBase
from ..subreddit.moderation import ModAction
from ..subreddit.subreddit import Subreddit
from ..subreddit.wiki import WikipageRevision

if TYPE_CHECKING:
    from ...reddit import Reddit


class Listing(aPRAWBase, Iterator):
    """
    A model representing Reddit listings.

    Members
    -------
    reddit: Reddit
        The :class:`~apraw.Reddit` instance with which requests are made.
    data: Dict
        The data retrieved from the endpoint.
    kind_filter:
        Kinds to return if given, otherwise all are returned.
    subreddit: Subreddit
        The subreddit to inject into items as their owner.
    CHILD_ATTRIBUTE: str
        The attribute in the data that contains the listing's items.
    """

    CHILD_ATTRIBUTE = "children"

    def __init__(self, reddit: 'Reddit', data: Dict, kind_filter: List[str] = None,
                 subreddit: Subreddit = None, link_id: str = ""):
        """
        Create a ``Listing`` instance.

        Parameters
        ----------
        reddit: Reddit
            The :class:`~apraw.Reddit` instance with which requests are made.
        data: Dict
            The data retrieved from the endpoint.
        kind_filter:
            Kinds to return if given, otherwise all are returned.
        subreddit: Subreddit
            The subreddit to inject into items as their owner.
        """
        super().__init__(reddit, data, reddit.listing_kind)

        self._index = 0
        self._subreddit = subreddit
        self._link_id = link_id
        self._kind_filter = kind_filter if kind_filter else []

    def __len__(self) -> int:
        """
        Return the number of items in the Listing.

        Returns
        -------
        len: int
            The number of items in the listing.
        """
        return len(getattr(self, self.CHILD_ATTRIBUTE))

    def __iter__(self) -> Iterator[aPRAWBase]:
        """
        Permit Listing to operate as an iterator.

        Returns
        -------
        self: Listing
            The iterator.
        """
        return self

    def __next__(self) -> aPRAWBase:
        """
        Permit Listing to operate as a generator.

        Returns
        -------
        item: aPRAWBase
            The next item in the listing.
        """
        while True:
            if self._index >= len(self):
                raise StopIteration()
            self._index += 1
            item = self[self._index - 1]

            if not self._kind_filter or (self._kind_filter and item.kind in self._kind_filter):
                break

        return item

    def __getitem__(self, index: int) -> aPRAWBase:
        """
        Return the item at position index in the list.

        Parameters
        ----------
        index : int
            The item's index.

        Returns
        -------
        item: aPRAWBase
            The searched item.
        """
        item = getattr(self, self.CHILD_ATTRIBUTE)[index]

        if isinstance(item, aPRAWBase):
            return item

        if "page" in item:
            return WikipageRevision(self._reddit, item)
        elif item["kind"] == self._reddit.link_kind:
            return Submission(self._reddit, item["data"], subreddit=self._subreddit)
        elif item["kind"] == self._reddit.subreddit_kind:
            return Subreddit(self._reddit, item["data"])
        elif item["kind"] == self._reddit.comment_kind:
            if item["data"]["replies"] and item["data"]["replies"]["kind"] == self._reddit.listing_kind:
                from ..helpers.comment_forrest import CommentForrest
                replies = CommentForrest(self._reddit, item["data"]["replies"]["data"], item["data"]["link_id"])
            else:
                replies = []
            return Comment(self._reddit, item["data"], subreddit=self._subreddit, replies=replies)
        elif item["kind"] == self._reddit.modaction_kind:
            return ModAction(self._reddit, item["data"], self._subreddit)
        elif item["kind"] == self._reddit.message_kind:
            return Message(self._reddit, item["data"])
        elif item["kind"] == self._reddit.listing_kind:
            return Listing(self._reddit, item["data"])
        elif item["kind"] == self._reddit.more_kind:
            return MoreComments(self._reddit, item["data"], self._link_id)
        else:
            return aPRAWBase(self._reddit, item["data"] if "data" in item else item)

    @property
    def last(self) -> aPRAWBase:
        """
        Return the last item in the listing.

        Returns
        -------
        item: aPRAWBase
            The last item in the listing.
        """
        return self[len(self) - 1] if len(self) > 0 else None


class MoreChildren(Listing):
    CHILD_ATTRIBUTE = "things"



import asyncio
from datetime import datetime
from typing import TYPE_CHECKING, Dict, List, Any, Union

from reactive import all_reactive, ReactiveOwner

from .redditor import Redditor
from ..helpers.apraw_base import aPRAWBase
from ..helpers.item_moderation import PostModeration
from ..mixins.author import AuthorMixin
from ..mixins.deletable import DeletableMixin
from ..mixins.hideable import HideableMixin
from ..mixins.replyable import ReplyableMixin
from ..mixins.savable import SavableMixin
from ..mixins.subreddit import SubredditMixin
from ..mixins.votable import VotableMixin
from ..subreddit.subreddit import Subreddit
from ...const import API_PATH
from ...utils import prepend_kind, snake_case_keys, ExponentialCounter

if TYPE_CHECKING:
    from ...reddit import Reddit
    from .submission import Submission
    from ..helpers.comment_forrest import CommentForrest


@all_reactive(not_type=(aPRAWBase, datetime, PostModeration))
class Comment(aPRAWBase, DeletableMixin, HideableMixin, ReplyableMixin, SavableMixin, VotableMixin, AuthorMixin,
              SubredditMixin, ReactiveOwner):
    """
    The model representing comments.

    Members
    -------
    mod: CommentModeration
        The :class:`~apraw.models.CommentModeration` instance to aid in moderating the comment.
    kind: str
        The item's kind / type.
    url: str
        The URL pointing to this comment.

    **Typical Attributes**

    This table describes attributes that typically belong to objects of this
    class. Attributes are dynamically provided by the :class:`~apraw.models.aPRAWBase` class
    and may vary depending on the status of the response and expected objects.

    ================================= ===========================================================================
    Attribute                         Description
    ================================= ===========================================================================
    ``all_awardings``                 A list of awardings added to the comment.
    ``approved_at_utc``               The UTC timestamp at which the comment was approved by the moderators.
    ``approved_by``                   The moderator who approved this comment if applicable.
    ``approved``                      Whether the comment has been approved by the moderators.
    ``archived``                      Whether the comment has been archived.
    ``author_flair_background_color`` The comment author's flair background color if applicable.
    ``author_flair_css_class``        The comment author's flair CSS class if applicable.
    ``author_flair_richtext``         The comment author's flair text if applicable.
    ``author_flair_template_id``      The comment author's flair template ID if applicable.
    ``author_flair_text_color``       The comment author's flair text color if applicable.
    ``author_flair_text``             The comment author's flair text if applicable.
    ``author_flair_type``             The comment author's flair type if applicable.
    ``author_fullname``               The comment author's ID prepended with ``t2_``.
    ``author_patreon_flair``          The comment author's Patreon flair if applicable.
    ``author``                        The comment author's username.
    ``banned_at_utc``                 ``None``
    ``banned_by``                     ``None``
    ``body_html``                     The HTML version of the comment's body.
    ``body``                          The comment's markdown body.
    ``can_gild``                      Whether the logged-in user can gild the comment.
    ``can_mod_post``                  ``bool``
    ``collapsed_reason``              ``None``
    ``collapsed``                     Whether the comment should be collapsed by clients.
    ``controversiality``              A score on the comment's controversiality based on its up- and downvotes.
    ``created_utc``                   The parsed UTC ``datetime`` on which the comment was made.
    ``created``                       A timestamp on which the comment was created.
    ``distinguished``                 The type of distinguishment the comment hsa received.
    ``downs``                         The number of downvotes the comment has received.
    ``edited``                        Whether the comment has been edited from its original state.
    ``gilded``                        The number of awards this comment has received.
    ``gildings``                      A dictionary of gilds the comment has received.
    ``id``                            The comment's ID.
    ``ignore_reports``                Whether reports should be ignored on this comment.
    ``is_submitter``                  Whether the logged-in user is the submitter of this comment.
    ``likes``                         The overall upvote score on this comment.
    ``link_author``                   The username of the comment submission's author.
    ``link_id``                       The ID of the submission this comment was made in.
    ``link_permalink`` / ``link_url`` A URL to the comment's submission.
    ``link_title``                    The comment's submission title.
    ``locked``                        Whether the comment has been locked by the moderators.
    ``mod_note``                      Notes added to the comment by moderators if applicable.
    ``mod_reason_by``                 The moderator who added a removal reason if applicable.
    ``mod_reason_title``              The mod reason's title if applicable.
    ``mod_reports``                   A list of reports made on this comment filed by moderators.
    ``name``                          The comment's ID prepended with ``t1_``.
    ``no_follow``                     ``bool``
    ``num_comments``                  The number of replies made in this submission.
    ``num_reports``                   The number of reports on this comment.
    ``over_18``                       Whether the comment has been marked NSFW.
    ``parent_id``                     The comment's parent ID, either ``link_id`` or the ID of another comment.
    ``permalink``                     The comment's permalink.
    ``quarantine``                    ``bool``
    ``removal_reason``                A removal reason set by moderators if applicable.
    ``removed``                       Whether the comment has been removed by the moderators of the subreddit.
    ``replies``                       A list of replies made under this comment, usually empty at first.
    ``report_reasons``                Report reasons added to the comment.
    ``saved``                         Whether the logged-in user has saved this comment.
    ``score_hidden``                  Whether clients should hide the comment's score.
    ``score``                         The overall upvote score on this comment.
    ``send_replies``                  Whether the OP has enabled reply notifications.
    ``spam``                          Whether the comment has been flagged as spam.
    ``stickied``                      Whether the comment has been stickied by the moderators.
    ``subreddit_id``                  The comment subreddit's ID prepended with ``t5_``.
    ``subreddit_name_prefixed``       The comment's subreddit name prefixed with "r/".
    ``subreddit_type``                The type of the subreddit the submission was posted on
                                      (public, restricted, private).
    ``subreddit``                     The name of the subreddit this comment was made in.
    ``total_awards_received``         The number of awards this comment has received.
    ``ups``                           The number of upvotes this comment has received.
    ``user_reports``                  A list of user reports filed for this comment.
    ================================= ===========================================================================

    .. note::
        Many of these attributes are only available if the logged-in user has moderator access to the item.

    """

    def __init__(self, reddit: 'Reddit', data: Dict, submission: 'Submission' = None,
                 author: Redditor = None, subreddit: Subreddit = None, replies: Union['CommentForrest', List] = None):
        """
        Create an instance of a comment.

        Parameters
        ----------
        reddit: Reddit
            The :class:`~apraw.Reddit` instance with which requests are made.
        data: Dict
            The data obtained from the /about endpoint.
        submission: Submission
            The submission this comment was made in as a :class:`~apraw.models.Submission`.
        author: Redditor
            The author of this comment as a :class:`~apraw.models.Redditor`.
        subreddit: Subreddit
            The subreddit this comment was made in as a :class:`~apraw.models.Subreddit`.
        replies: List[Comment]
            A list of replies made to this comment.
        """
        self.replies = replies if replies else []

        ReactiveOwner.__init__(self)
        aPRAWBase.__init__(self, reddit, data, reddit.comment_kind)
        AuthorMixin.__init__(self, author)
        SubredditMixin.__init__(self, subreddit)

        self._submission = submission
        self.mod = CommentModeration(reddit, self)

    async def fetch(self):
        """
        Fetch this item's information from a suitable API endpoint.

        Returns
        -------
        self: Comment
            The ``Comment`` model with updated data.
        """
        if ("link_id" in self._data and "id" in self._data and "subreddit" in self._data) or "permalink" in self._data:
            permalink = self._data["permalink"] if "permalink" in self._data else API_PATH["comment"].format(
                sub=self._data["subreddit"], submission=self._data["link_id"].replace(self._reddit.link_kind + "_", ""),
                id=self._data["id"])
            resp = await self._reddit.get_request(permalink)
            from .submission import Submission
            self._submission = Submission(self._reddit, resp[0]["data"]["children"][0]["data"])
            return self._update(resp[1]["data"]["children"][0]["data"])
        elif "id" in self._data:
            resp = await self._reddit.get_request(API_PATH["info"],
                                                  id=prepend_kind(self._data["id"], self._reddit.comment_kind))
            return self._update(resp["data"]["children"][0]["data"])
        else:
            raise ValueError(f"No data available to make request URL: {self._data}")

    async def monitor(self, max_wait=16):
        counter = ExponentialCounter(max_wait)
        while True:
            updates = await self.fetch()
            if updates:
                wait = counter.reset()
            else:
                wait = counter.count()
            await asyncio.sleep(wait)

    def _update(self, _data: Union[List, Dict[str, Any]]):
        """
        Update the base with new information.

        Parameters
        ----------
        _data: Dict
            The data obtained from the API.
        """

        if isinstance(_data, dict) or isinstance(_data, list):
            data = _data if isinstance(_data, dict) else _data[0]["data"]
            if isinstance(_data, list):
                from ..helpers.comment_forrest import CommentForrest
                self.replies = CommentForrest(self._reddit, _data[1]["data"], data["link_id"])

            self._data = data

            d = snake_case_keys(data)
            if "created_utc" in d:
                d["created_utc"] = datetime.utcfromtimestamp(d["created_utc"])
            self._data_attrs.update([k for k in d if not hasattr(self, k)])
            updates = [{"name": k, "value": v} for (k, v) in d.items() if not hasattr(self, k) or k in self._data_attrs]

            if "link_id" in d and "id" in d and "subreddit" in d:
                link_id = d["link_id"].replace(self._reddit.link_kind + "_", "")
                url = API_PATH["comment"].format(sub=d["subreddit"], submission=link_id, id=d["id"])
                updates.append({"name": "url", "value": url})
            elif "permalink" in d:
                url = "https://www.reddit.com" + d["permalink"]
                updates.append({"name": "url", "value": url})

            return self._bulk_update(*updates)
        else:
            raise TypeError("data is not of type 'dict' or 'list'.")

    async def submission(self) -> 'Submission':
        """
        Retrieve the submission this comment was made in as a :class:`~apraw.models.Submission`.

        Returns
        -------
        submission: Submission
            The submission this comment was made in.
        """
        if self._submission is None:
            link = await self._reddit.get_request(API_PATH["info"], id=self._data["link_id"])
            from .submission import Submission
            self._submission = Submission(
                self._reddit, link["data"]["children"][0]["data"])
        return self._submission


class CommentModeration(PostModeration):
    """
    A helper class to moderate comments.

    Members
    -------
    reddit: Reddit
        The :class:`~apraw.Reddit` instance with which requests are made.
    fullname: str
        The ID prepended with the kind of the item this helper belongs to.
    """

    def __init__(self, reddit: 'Reddit', comment: Comment):
        """
        Create an instance of ``CommentModeration``.

        Parameters
        ----------
        reddit : Reddit
            The :class:`~apraw.Reddit` instance with which requests are made.
        comment : Comment
            The comment this helper performs requests for.
        """
        super().__init__(reddit, comment)

    async def show_comment(self):
        """
        Mark a comment that it should not be collapsed because of crowd control.

        The comment could still be collapsed for other reasons.

        Returns
        -------
        resp: Dict
            The API response JSON.
        """
        return await self.reddit.post_request(API_PATH["mod_show_comment"], id=self.fullname)



from datetime import datetime
from typing import TYPE_CHECKING, Dict, List

import aiohttp

from .helpers.apraw_base import aPRAWBase
from .helpers.generator import ListingGenerator
from .helpers.streamable import Streamable
from .reddit.redditor import Redditor
from ..endpoints import API_PATH

if TYPE_CHECKING:
    from ..reddit import Reddit


class User:
    """
    A class to store the authentication credentials and handle ratelimit information.

    Members
    -------
    reddit: Reddit
        The :class:`~apraw.Reddit` instance with which requests are made.
    username: str
        The username given to the Reddit instance or obtained via ``praw.ini``.
    password: str
        The password given to the Reddit instance or obtained via ``praw.ini``.
    client_id: str
        The client ID given to the Reddit instance or obtained via ``praw.ini``.
    client_secret: str
        The client secret given to the Reddit instance or obtained via ``praw.ini``.
    user_agent: str
        The user agent given to the Reddit instance or defaulted to aPRAW's version.
    password_grant: str
        The data to be used when making a token request with the 'password' ``grant_type``.
    access_data: Dict
        A dictionary containing the access token and user agent for request headers.
    token_expires: datetime
        The datetime on which the previously retrieved token will expire. Defaults to the past to obtain a token
        immediately the first time.
    ratelimit_remaining: int
        The number of requests remaining in the current ratelimit window.
    ratelimit_used: int
        The number of requests previously used in the current ratelimit window.
    ratelimit_reset: datetime
        The datetime on which the ratelimit window will be reset.
    """

    def __init__(self, reddit: 'Reddit', username: str, password: str, client_id: str,
                 client_secret: str, user_agent: str):
        """
        Create an instance of the authenticated user.

        Parameters
        ----------
        reddit: Reddit
            The :class:`~apraw.Reddit` instance with which requests are made.
        username: str
            The username given to the Reddit instance or obtained via ``praw.ini``.
        password: str
            The password given to the Reddit instance or obtained via ``praw.ini``.
        client_id: str
            The client ID given to the Reddit instance or obtained via ``praw.ini``.
        client_secret: str
            The client secret given to the Reddit instance or obtained via ``praw.ini``.
        user_agent: str
            The user agent given to the Reddit instance or defaulted to aPRAW's version.

        Raises
        ------
        Exception
            If the login credentials given are empty or incomplete.
        """
        self.reddit = reddit

        self.username = username
        self.password = password
        self.client_id = client_id
        self.client_secret = client_secret
        self.user_agent = user_agent

        if self.username == "" or self.password == "" or self.client_id == "" or self.client_secret == "":
            raise Exception(
                "No login information given or login information incomplete.")

        self.password_grant = "grant_type=password&username={}&password={}".format(
            self.username, self.password)

        self._auth_session = None
        self._client_session = None

        self._auth_user = None

        self.access_data = None
        self.token_expires = datetime.now()

        self.ratelimit_remaining = 0
        self.ratelimit_used = 0
        self.ratelimit_reset = datetime.now()

    async def auth_session(self) -> aiohttp.ClientSession:
        """
        Retrieve an ``aiohttp.ClientSesssion`` with which the authentication token can be obtained.

        Returns
        -------
        session: aiohttp.ClientSession
            The session using the BasicAuth setup to obtain tokens with.
        """
        if self._auth_session is None:
            auth = aiohttp.BasicAuth(
                login=self.client_id,
                password=self.client_secret)
            self._auth_session = aiohttp.ClientSession(auth=auth)
        return self._auth_session

    async def client_session(self) -> aiohttp.ClientSession:
        """
        Retrieve the ``aiohttp.ClientSesssion`` with which regular requests are made.

        Returns
        -------
        session: aiohttp.ClientSession
            The session with which requests should be made.
        """
        if self._client_session is None:
            self._client_session = aiohttp.ClientSession()
        return self._client_session

    async def close(self):
        await self.auth_session.close()
        await self.client_session.close()

    async def me(self) -> 'AuthenticatedUser':
        """
        Retrieve an instance of :class:`~apraw.models.AuthenticatedUser` for the logged-in user.

        Returns
        -------
        user: AuthenticatedUser
            The logged-in user.
        """
        if not self._auth_user:
            data = await self.reddit.get_request(API_PATH["me"])
            self._auth_user = AuthenticatedUser(self.reddit, data)
        return self._auth_user


class AuthenticatedUser(Redditor):
    """
    The model representing the logged-in user.

    This model inherits from :class:`~apraw.models.Redditor` and thus all its attributes and features. View those docs
    for further information.

    Members
    -------
    reddit: Reddit
        The :class:`~apraw.Reddit` instance with which requests are made.
    data: Dict
        The data obtained from the /about endpoint.
    """

    def __init__(self, reddit: 'Reddit', data: Dict):
        """
        Create an instance of AuthenticatedUser.

        Parameters
        ----------
        reddit : Reddit
            The :class:`~apraw.Reddit` instance with which requests are made.
        data : Dict
            The data obtained from the /about endpoint.
        """
        super().__init__(reddit, data)

        self._karma = list()

    async def karma(self) -> List['Karma']:
        """
        Retrieve the karma breakdown for the logged-in user.

        Returns
        -------
        karma: List[Karma]
            The parsed ``KarmaList`` for the logged-in user.
        """
        if not self._karma:
            resp = await self.reddit.get_request(API_PATH["me_karma"])
            self._karma = [Karma(self.reddit, d) for d in resp["data"]]
        return self._karma

    @Streamable.streamable
    async def inbox(self, *args, **kwargs) -> ListingGenerator:
        return ListingGenerator(self.reddit, API_PATH["message_inbox"], *args, **kwargs)

    @Streamable.streamable
    async def sent(self, *args, **kwargs) -> ListingGenerator:
        return ListingGenerator(self.reddit, API_PATH["message_sent"], *args, **kwargs)

    @Streamable.streamable
    async def unread(self, *args, **kwargs) -> ListingGenerator:
        return ListingGenerator(self.reddit, API_PATH["message_unread"], *args, **kwargs)


class Karma(aPRAWBase):
    """
    A model representing subreddit karma.

    Members
    -------
    reddit: Reddit
        The :class:`~apraw.Reddit` instance with which requests are made.
    data: Dict
        The data obtained from the /about endpoint.

    **Typical Attributes**

    This table describes attributes that typically belong to objects of this
    class. Attributes are dynamically provided by the :class:`~apraw.models.aPRAWBase` class
    and may vary depending on the status of the response and expected objects.

    ================= ===================================================
    Attribute         Description
    ================= ===================================================
    ``sr``            The name of the subreddit the karma was obtained on
    ``comment_karma`` The amount of karma obtained on the subreddit.
    ``link_karma``    The amount of link karma obtained on the subreddit.
    ================= ===================================================
    """

    def __init__(self, reddit: 'Reddit', data: Dict):
        """
        Create an instance of Karma

        Parameters
        ----------
        reddit : Reddit
            An instance of :class:`~apraw.Reddit`with which requests are made.
        data : Dict
            The data obtained from the /about endpoint.
        """
        super().__init__(reddit, data)

        self._subreddit = None

    async def subreddit(self):
        """
        Retrieve the subreddit on which the karma was obtained.

        Returns
        -------
        subreddit: Subreddit
            The subreddit on which the karma was obtained.
        """
        return await self._reddit.subreddit(self.sr)



from typing import TYPE_CHECKING, Dict, Any, List, Union

from .comment import Comment
from .submission import Submission
from ..helpers.apraw_base import aPRAWBase
from ...const import API_PATH

if TYPE_CHECKING:
    from ...reddit import Reddit


class MoreComments(aPRAWBase):

    def __init__(self, reddit: 'Reddit', data: Dict[str, Any], link_id: str):
        data.update(link_id=link_id)
        super().__init__(reddit, data)

        self._comments = []
        self._ids = list(self.children)
        self._index = 0

    async def _next_batch(self):
        if not self._ids:
            return

        ids = self._ids[:100]
        self._ids = self._ids[100:]
        resp = await self._reddit.get_request(API_PATH["morechildren"], **{
            "children": ",".join(ids),
            "link_id": self.link_id,
            "id": self.id,
            "depth": self.depth
        })

        from .listing import MoreChildren
        children = MoreChildren(self._reddit, resp["json"]["data"], [self._reddit.comment_kind, self._reddit.more_kind],
                                self.link_id)
        self._comments.extend(comment for comment in children)

    async def parent(self) -> Union[Submission, Comment]:
        return await self._reddit.info(self.parent_id)

    async def fetch(self):
        while self._ids:
            await self._next_batch()

    async def __anext__(self) -> Comment:
        if self._index >= len(self._comments) and not self._ids:
            raise StopAsyncIteration

        if not self._comments or self._ids:
            await self._next_batch()

        self._index += 1
        return self._comments[self._index + 1]

    async def comments(self) -> List[Comment]:
        if not self._comments:
            await self.fetch()

        return self._comments



import re
from typing import Any, Dict

pattern = re.compile(r'(?<!^)(?=[A-Z])')


def camel_to_snake(name: str) -> str:
    return pattern.sub("_", name).lower()


def snake_case_keys(dictionary: Dict[str, Any]) -> Dict[str, Any]:
    return {camel_to_snake(k): v for k, v in dictionary.items()}



from ...const import API_PATH


class DeletableMixin:
    """
    Mixin for deletable objects.
    """

    async def delete(self):
        """
        Delete the item.

        Returns
        -------
        resp: Dict
            The API response JSON.
        """
        return await self._reddit.post_request(API_PATH["post_delete"], id=self.fullname)



"""List of Reddit API endpoints known to aPRAW."""
from typing import Dict

BASE_URL = "https://oauth.reddit.com{}?{}"

API_PATH: Dict[str, str] = {
    "comment"              : "/r/{sub}/comments/{submission}/_/{id}",
    "compose"              : "/api/compose",
    "info"                 : "/api/info",
    "me"                   : "/api/v1/me",
    "me_karma"             : "/api/v1/me/karma",
    "message_inbox"        : "/message/inbox",
    "message_sent"         : "/message/sent",
    "message_unread"       : "/message/unread",
    "mod_approve"          : "/api/approve",
    "mod_distinguish"      : "/api/distinguish",
    "mod_ignore_reports"   : "/api/ignore_reports",
    "mod_lock"             : "/api/lock",
    "mod_remove"           : "/api/remove",
    "mod_show_comment"     : "/api/show_comment",
    "mod_sticky"           : "/api/set_subreddit_sticky",
    "mod_unignore_reports" : "/api/unignore_reports",
    "mod_unlock"           : "/api/unlock",
    "moderated"            : "/user/{user}/moderated_subreddits",
    "modmail_conversation" : "/api/mod/conversations/{id}",
    "modmail_conversations": "/api/mod/conversations",
    "morechildren"         : "/api/morechildren",
    "post_delete"          : "/api/del",
    "post_hide"            : "/api/hide",
    "post_marknsfw"        : "/api/marknsfw",
    "post_save"            : "/api/save",
    "post_spoiler"         : "/api/spoiler",
    "post_unhide"          : "/api/unhide",
    "post_unmarknsfw"      : "/api/unmarknsfw",
    "post_unsave"          : "/api/unsave",
    "post_unspoiler"       : "/api/unspoiler",
    "post_vote"            : "/api/vote",
    "quarantine_opt_in"    : "/api/quarantine_optin",
    "quarantine_opt_out"   : "/api/quarantine_optout",
    "reply"                : "/api/comment",
    "submission"           : "/r/{sub}/comments/{id}",
    "subreddit"            : "/r/{sub}",
    "subreddit_about"      : "/r/{sub}/about",
    "subreddit_comments"   : "/r/{sub}/comments",
    "subreddit_edited"     : "/r/{sub}/about/edited",
    "subreddit_hot"        : "/r/{sub}/hot",
    "subreddit_log"        : "/r/{sub}/about/log",
    "subreddit_moderators" : "/r/{sub}/about/moderators",
    "subreddit_modqueue"   : "/r/{sub}/about/modqueue",
    "subreddit_new"        : "/r/{sub}/new",
    "subreddit_random"     : "/r/{sub}/random",
    "subreddit_reports"    : "/r/{sub}/about/reports",
    "subreddit_rising"     : "/r/{sub}/rising",
    "subreddit_spam"       : "/r/{sub}/about/spam",
    "subreddit_top"        : "/r/{sub}/top",
    "subreddit_unmoderated": "/r/{sub}/about/unmoderated",
    "subreddits_new"       : "/subreddits/new",
    "user_about"           : "/user/{user}/about",
    "user_comments"        : "/user/{user}/comments",
    "user_submissions"     : "/user/{user}/submitted",
    "wiki"                 : "/r/{sub}/wiki/pages",
    "wiki_alloweditor"     : "/r/{sub}/api/wiki/alloweditor/{act}",
    "wiki_edit"            : "/r/{sub}/api/wiki/edit",
    "wiki_hide"            : "/r/{sub}/api/wiki/hide",
    "wiki_page"            : "/r/{sub}/wiki/{page}",
    "wiki_page_revisions"  : "/r/{sub}/wiki/revisions/{page}",
    "wiki_revert"          : "/r/{sub}/api/wiki/revert",
    "wiki_revisions"       : "/r/{sub}/wiki/revisions",
    "submit"               : "/api/submit",
}



import setuptools
from apraw import __version__, __tag__

with open("README.md", "r", encoding="utf8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="aPRAW",
    version="{}-{}".format(__version__, __tag__) if __tag__ else __version__,
    author="Dan6erbond",
    author_email="moravrav@gmail.com",
    description="aPRAW is an asynchronous Reddit API wrapper written in Python.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Dan6erbond/aPRAW",
    packages=setuptools.find_packages(include=['apraw', 'apraw.*']),
    install_requires=[
        'aiohttp>=3.6.2',
        'reactivepy>=1.9.0.dev0'
    ],
    keywords="reddit api wrapper async",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Education",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content :: Message Boards",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
        "Typing :: Typed"
    ],
    license="GNU General Public License v3 (GPLv3)",
    python_requires='>=3.6',
)

# classifiers can be found here: https://pypi.org/classifiers/



from typing import TYPE_CHECKING, Dict, List, Union

from ..helpers.apraw_base import aPRAWBase
from ..helpers.streamable import Streamable
from ..reddit.redditor import Redditor
from ...const import API_PATH

if TYPE_CHECKING:
    from ...reddit import Reddit
    from .subreddit import Subreddit


class SubredditWiki:

    def __init__(self, subreddit: 'Subreddit'):
        self.subreddit = subreddit
        self._data = None

    @Streamable.streamable
    def revisions(self, *args, **kwargs):
        r"""
        Returns an instance of :class:`~apraw.models.ListingGenerator` mapped to recent wikipage revisions.

        .. note::
            This listing can be streamed doing the following:

            .. code-block:: python3

                for comment in subreddit.wiki.stream():
                    print(comment)

        Parameters
        ----------
        kwargs: \*\*Dict
            :class:`~apraw.models.ListingGenerator` ``kwargs``.

        Returns
        -------
        generator: ListingGenerator
            A :class:`~apraw.models.ListingGenerator` mapped to recent wikipage revisions.
        """
        from ..helpers.generator import ListingGenerator
        return ListingGenerator(self.subreddit._reddit, API_PATH["wiki_revisions"].format(sub=self.subreddit), *args,
                                **kwargs)

    async def data(self, refresh=False) -> Dict:
        if self._data is None:
            self._data = await self.subreddit._reddit.get_request(
                API_PATH["wiki"].format(sub=self.subreddit))
        return self._data

    async def __call__(self) -> List[str]:
        data = await self.data()
        return [page for page in data["data"]]

    async def page(self, page: str) -> 'SubredditWikipage':
        resp = await self.subreddit._reddit.get_request(
            API_PATH["wiki_page"].format(sub=self.subreddit, page=page))
        return SubredditWikipage(page, self.subreddit, resp["data"])

    async def create(self, page: str, content_md: str = "", reason: str = "") -> 'SubredditWikipage':
        resp = await self.subreddit._reddit.post_request(
            API_PATH["wiki_edit"].format(sub=self.subreddit), data={
                "page": page,
                "content": content_md,
                "reason": reason
            })
        return resp if resp else await self.page(page)


class SubredditWikipage(aPRAWBase):

    def __init__(self, name: str, subreddit: 'Subreddit', data: Dict = None):
        super().__init__(subreddit._reddit, data, subreddit._reddit.wikipage_kind)

        self.name = name
        self.subreddit = subreddit

    @Streamable.streamable
    def revisions(self, *args, **kwargs):
        r"""
        Returns an instance of :class:`~apraw.models.ListingGenerator` mapped to fetch specific wikipage revisions.

        .. note::
            This listing can be streamed doing the following:

            .. code-block:: python3

                for comment in subreddit.wiki.page("test").stream():
                    print(comment)

        Parameters
        ----------
        kwargs: \*\*Dict
            :class:`~apraw.models.ListingGenerator` ``kwargs``.

        Returns
        -------
        generator: ListingGenerator
            A :class:`~apraw.models.ListingGenerator` mapped to fetch specific wikipage revisions.
        """
        from ..helpers.generator import ListingGenerator
        return ListingGenerator(self.subreddit._reddit, API_PATH["wiki_page_revisions"].format(
            sub=self.subreddit, page=self.name), *args, **kwargs)

    async def _alloweditor(self, username: str, act: str):
        resp = await self.subreddit._reddit.post_request(
            API_PATH["wiki_alloweditor"].format(sub=self.subreddit, act=act), data={
                "page": self.name,
                "username": username
            })
        return True if not resp else resp

    async def add_editor(self, username: str):
        return await self._alloweditor(username, "add")

    async def del_editor(self, username: str):
        return await self._alloweditor(username, "del")

    async def edit(self, content_md: str = "", reason: str = "") -> bool:
        resp = await self.subreddit._reddit.post_request(
            API_PATH["wiki_edit"].format(sub=self.subreddit), data={
                "page": self.name,
                "content": content_md,
                "reason": reason
            })
        return resp if resp else True

    async def hide(self, revision: Union[str, 'WikipageRevision']):
        resp = await self.subreddit._reddit.post_request(
            API_PATH["wiki_hide"].format(sub=self.subreddit), data={
                "page": self.name,
                "revision": str(revision)
            })
        return resp if resp else True

    async def revert(self, revision: Union[str, 'WikipageRevision']):
        resp = await self.subreddit._reddit.post_request(
            API_PATH["wiki_revert"].format(sub=self.subreddit), data={
                "page": self.name,
                "revision": str(revision)
            })
        return resp if resp else True


class WikipageRevision(aPRAWBase):

    def __init__(self, reddit: 'Reddit', data: Dict = None):
        super().__init__(reddit, data, reddit.wiki_revision_kind)

        self.author = Redditor(reddit, data["author"]["data"])

    def __str__(self):
        return self.id



from typing import TYPE_CHECKING, Union

from ...endpoints import API_PATH

if TYPE_CHECKING:
    from ..reddit.message import Message
    from ..reddit.comment import Comment


class ReplyableMixin:
    """
    Mixin for replyable objects.
    """

    async def reply(self, text: str) -> Union['Comment', 'Message']:
        """
        Reply to the item.

        Returns
        -------
        reply: Comment or Message
            The newly created reply, either a :class:`~apraw.models.Comment` or :class:`~apraw.models.Message`.
        """
        resp = await self._reddit.post_request(API_PATH["reply"], text=text, return_rtjson=True, thing_id=self.fullname)

        if self.kind == self._reddit.message_kind:
            from ..reddit.message import Message
            reply = Message(self._reddit, resp)
        else:
            from ..reddit.comment import Comment
            reply = Comment(self._reddit, resp)

        return reply

    comment = reply



from ...const import API_PATH


class VotableMixin:
    """
    Mixin for votable objects.
    """

    async def upvote(self):
        """
        Upvote the item.

        Returns
        -------
        resp: Dict
            The API response JSON.
        """
        return await self.reddit.post_request(API_PATH["post_vote"], id=self.fullname, dir=1)

    async def downvote(self):
        """
        Downvote the item.

        Returns
        -------
        resp: Dict
            The API response JSON.
        """
        return await self.reddit.post_request(API_PATH["post_vote"], id=self.fullname, dir=-1)

    async def clear_vote(self):
        """
        Clear user up- and downvotes on the item.

        Returns
        -------
        resp: Dict
            The API response JSON.
        """
        return await self.reddit.post_request(API_PATH["post_vote"], id=self.fullname, dir=0)
